from django.urls import path, include
from rest_framework.routers import DefaultRouter
from .views.patient_views import (
    PatientViewSet,
    start_examination,
    get_location_choices,
    get_morphology_choices, 
)
from .views.csrf import csrf_token_view
#from .views.feature_selection_view import FetchSingleFramePredictionView // its implemented in ando-ai other project need to add here
from .views.video_segmentation_views import VideoView, VideoLabelView,UpdateLabelSegmentsView
from .views.views_for_timeline import video_timeline_view
from .views.raw_video_meta_validation_views import VideoFileForMetaView
router = DefaultRouter()
router.register(r'patients', PatientViewSet)

urlpatterns = [
    path('start-examination/', start_examination, name="start_examination"),
    path('get-location-choices/<int:location_id>/', get_location_choices, name="get_location_choices"),
    path('get-morphology-choices/<int:morphology_id>/', get_morphology_choices, name="get_morphology_choices"),
    path('api/', include(router.urls)),
    path('api/conf/', csrf_token_view, name='csrf_token'),
    


#--------------------------------------VIDEO SEGMENTATION END POINTS--------------------------------------

       # The dropdown contains video names and their corresponding IDs, which are retrieved from the database(RawVideoFile). Additionally, this route(api/videos) also fetches labels along with their names and IDs from the label table.
       # We will modify this implementation later as per our requirements.
        #Based on the selected video ID, relevant data will be fetched from the database, and the video, along with its associated details, 
        # will be displayed on the frontend using route(api/video/<int:video_id>/).
        # once label is selected from the dropdown,using its name, details can be fetched from rawvideofile using route("api/video/<int:video_id>/label/<str:label_name>/) 
        #If editing is required, a form will be available for each label. This form dynamically updates when the selected label changes. It will display all segments associated with the selected label, each with a delete option. Below these segments, there may be a button for adding more segments.
        #If any values in the form are modified, the updated data will be saved in the database table.
    
    
    
    
    # Fetch all videos for the dropdown
    # Fetch all available videos (for the video selection dropdown in frontend)
    # Usage in frontend: Call this API to get a list of available videos.
    # Example request: GET /api/videos/
    # Response: Returns a JSON list of videos (id, file path, video_url, sequences, label_names).
    # it also fetch the label from teh label table
    path("api/videos/", VideoView.as_view(), name="video_list"),
    
    #need to change this route
    #path('api/prediction/', FetchSingleFramePredictionView.as_view(), name='fetch-single-frame-prediction'),#.as_view() converts the class-based view into a function that Django can use for routing.
    
    # Fetch video details + labels based on selected video
    # Usage in frontend: Call this when a user selects a video from the dropdown.
    # Example request: GET /api/video/1/
    # Response: Returns metadata, file path, video URL, sequences, and available label names.
    # Frontend should use this api for `label_names`, to display label selection (dropdown or tabs).
    path("api/video/<int:video_id>/", VideoView.as_view(), name="video_handler"),
    

    # Fetch time segments along with frame predictions and filenames for a selected label
    # 
    # **Usage in frontend:**
    # - Call this API when a user selects a label from the dropdown in the timeline.
    # - The API returns time segments (start & end times in seconds) where the label appears in the video.
    # - Each segment includes frame-wise predictions and the corresponding frame filenames.
    #
    # **Example request:**
    #   GET /api/video/1/label/outside/
    #
    # **Example response:**
    # {
    #     "label": "outside",
    #     "time_segments": [
    #         {   "segment_start": 0,
     #       "segment_end": 463,
    #             "start_time": 0.0,
    #             "end_time": 9.26,
    #             "frames": {
    #                 "0": {
    #                     "frame_filename": "frame_0000000.jpg",
    #                     "frame_file_path": "/home/admin/test-data/video_frames/abc123/frame_0000000.jpg",
    #                     "predictions": {
    #                         "appendix": 0.0150,
    #                         and others
    #                     }
    #                 },
    #                 "1": {
    #                     "frame_filename": "frame_0000001.jpg",
    #                     "frame_file_path": "/home/admin/test-data/video_frames/abc123/frame_0000001.jpg",
    #                     "predictions": {
    #                         "appendix": 0.0124,
    #                         and otehrs
    #                     }}}}]}
    #                 
    # start_time and end_time are in seconds, converted from frame indices.
    # frames contains all frames in the segment with:
    # frame_filename (e.g., frame_0000001.jpg)
    # frame_file_path (full path for easy access)
    #  predictions (frame-wise probability scores for each category)
    #we need label name for this not id as we donot have id in rawvideofile table
    path("api/video/<int:video_id>/label/<str:label_name>/", VideoLabelView.as_view(), name="video_label_times"),

    # Update or create label segments for a specific video and label
    # - Call this API when a user updates the start and end frames of a label segment(adding new , deleting or changing the values).
    # - The API will:
    #   - **Update** existing segments if they match the given `video_id` and `label_id`.
    #   - **Create** new segments if they do not already exist.
    #   - **Delete** segments from the database if they were removed from the frontend.
    # - This ensures that the database reflects exactly what is sent from the frontend.
    #
    # **Example request (Updating segments for label ID 12 in video ID 1):**
    #   PUT /api/video/1/label/12/update_segments/
    #
    # **Request body (JSON):**
    # {
    #     "video_id": 1,
    #     "label_id": 12,
    #     "segments": [
    #         {
    #             "start_frame_number": 0,
    #             "end_frame_number": 463
    #         },{
    #             "start_frame_number": 500,
    #             "end_frame_number": 700
    #         }]}
    #
    # **Expected response (If successful):**
    # {
    #     "updated_segments": [
    #         {
    #             "id": 1,
    #             "video_id": 1,
    #             "label_id": 12,
    #             "start_frame_number": 0,
    #             "end_frame_number": 463
    #         }
    #     ],
    #     "new_segments": [
    #         {
    #             "id": 2,
    #             "video_id": 1,
    #             "label_id": 12,
    #             "start_frame_number": 500,
    #             "end_frame_number": 700
    #         }
    #     ],
    #     "deleted_segments": 1  # Number of deleted segments if any were removed
    # }
 
    # **Frontend Integration:**
    # - Vue.js should send this request when the user **saves changes** to segments.
    # - The **"Add Segment"** button should allow users to add new segments dynamically.
    # - The **"Delete Segment"** button should remove segments from the list before submitting.
    # - The request should be triggered via an `axios.put` call:
    #
    path("api/video/<int:video_id>/label/<int:label_id>/update_segments/", UpdateLabelSegmentsView.as_view(), name="update_label_segments"),

#----------------------------------END--VIDEO SEGMENTATION SECTION-------------------------------
    #this is for to test the timeline
    #need to delete this url and also endoreg_db_production/endoreg_db/views/views_for_timeline.py and endoreg_db_production/endoreg_db/templates/timeline.html
    path('video/<int:video_id>/timeline/', video_timeline_view, name='video_timeline'),

    
    
    
    #  const url = lastId ? `http://localhost:8000/api/video/meta/?last_id=${lastId}` : "http://localhost:8000/api/video/meta/";
    path("api/video/sensitivemeta/", VideoFileForMetaView.as_view(), name="video_meta"),  # Single endpoint for both first and next video    
    ]






#https://biigle.de/manual/tutorials/videos/navigating-timeline#for time line example



"""Module for utility classes and functions."""

from .dataloader import load_model_data_from_yaml
from .parse_and_generate_yaml import collect_center_names
from .validate_endo_roi import validate_endo_roi
from .pydantic_models import DbConfig
from .dates import random_day_by_month_year, random_day_by_year
from .hashs import (
    get_hash_string,
    get_pdf_hash,
    get_examiner_hash,
    get_video_hash,
    get_patient_examination_hash,
)
from .names import (
    create_mock_patient_name,
    create_mock_examiner_name,
    guess_name_gender,
)

__all__ = [
    "load_model_data_from_yaml",
    "collect_center_names",
    "validate_endo_roi",
    "get_examiner_hash",
    "DbConfig",
    "random_day_by_month_year",
    "random_day_by_year",
    "get_hash_string",
    "create_mock_examiner_name",
    "create_mock_patient_name",
    "guess_name_gender",
    "get_pdf_hash",
    "get_video_hash",
    "get_patient_examination_hash",
]



"""Module for annotation models"""

from .image_classification import ImageClassificationAnnotation
from .binary_classification_annotation_task import (
    BinaryClassificationAnnotationTask,
)
from .anonymized_image_annotation import (
    AnonymousImageAnnotation,
    DroppedName,
    AnonymizedImageLabel,
    AnonymizedFile,
    UploadedFile,
)
from .video_segmentation_annotation import (
    VideoSegmentationLabel,
    VideoSegmentationAnnotation,
)

from .video_segmentation_labelset import VideoSegmentationLabelSet

__all__ = [
    "ImageClassificationAnnotation",
    "BinaryClassificationAnnotationTask",
    "AnonymousImageAnnotation",
    "DroppedName",
    "AnonymizedImageLabel",
    "AnonymizedFile",
    "UploadedFile",
    "VideoSegmentationLabel",
    "VideoSegmentationLabelSet",
    "VideoSegmentationAnnotation",
]



from django.db import models

class RuleTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class RuleType(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)

    objects = RuleTypeManager()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = 'Rule Type'
        verbose_name_plural = 'Rule Types'


"""
Django model for AI models.
"""

from pathlib import Path
import os
from django.db import models
from icecream import ic

PSEUDO_DIR = Path(os.environ.get("DJANGO_PSEUDO_DIR", Path("./erc_data")))

STORAGE_LOCATION = PSEUDO_DIR
WEIGHTS_DIR_NAME = "db_model_weights"
WEIGHTS_DIR = STORAGE_LOCATION / WEIGHTS_DIR_NAME

if not WEIGHTS_DIR.exists():
    WEIGHTS_DIR.mkdir(parents=True)


class AiModelManager(models.Manager):
    """
    Manager for AI models with custom query methods.
    """

    def get_by_natural_key(self, name: str) -> "AiModel":
        """
        Return the model with the given name.
        """
        return self.get(name=name)


class AiModel(models.Model):
    """
    Represents a generic AiModel.
    ModelMeta objects have a foreign key to this model.
    Here we gather high-level information about the model.
    this contains the name, description, labels, and version of the model.

    Attributes:
        name (str): The name of the model.
        name_de (str): The German name of the model.
        name_en (str): The English name of the model.
        description (str): A description of the model.
        model_type (str): The type of the model.
        model_subtype (str): The subtype of the model.
        video_segmentation_labels (QuerySet): The labels for video segmentation.
    """

    objects = AiModelManager()

    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    model_type = models.CharField(max_length=255, blank=True, null=True)
    model_subtype = models.CharField(max_length=255, blank=True, null=True)
    video_segmentation_labelset = models.ForeignKey(
        "VideoSegmentationLabelSet",
        on_delete=models.CASCADE,
        related_name="ai_models",
        blank=True,
        null=True,
    )
    active_meta = models.ForeignKey(
        "ModelMeta",
        on_delete=models.SET_NULL,
        related_name="active_model",
        blank=True,
        null=True,
    )

    @classmethod
    def set_active_model_meta(cls, model_name: str, meta_name: str, meta_version: int):
        """
        Set the active model.
        """
        from endoreg_db.models import ModelMeta

        model = cls.objects.get(name=model_name)
        assert model is not None, "Model not found"

        ic(f"Getting model meta for {model_name} {meta_name} {meta_version}")

        model_meta = ModelMeta.objects.get(
            name=meta_name, model=model, version=meta_version
        )
        assert model_meta is not None, "ModelMeta not found"

        model.active_meta = model_meta
        model.save()

        ic(
            f"Set active model meta for {model_name} to {meta_name} version {meta_version}"
        )

    def natural_key(self):
        """
        Return the natural key for this model.
        """
        return (self.name,)

    def __str__(self):
        return str(self.name)



""" """

from pathlib import Path
from collections import defaultdict, Counter
import shutil
import os
import subprocess
from typing import Optional, List, TYPE_CHECKING, Union
from django.db import models, transaction
from icecream import ic
from tqdm import tqdm
from endoreg_db.utils.hashs import get_video_hash
from endoreg_db.utils.file_operations import get_uuid_filename
from endoreg_db.utils.ocr import extract_text_from_rois

from ....utils.video import (
    transcode_videofile,
    transcode_videofile_if_required,
    initialize_frame_objects,
    extract_frames,
)

from ..metadata import VideoMeta, SensitiveMeta
from .utils import (
    STORAGE_LOCATION,
    VIDEO_DIR,
    FRAME_DIR,
)
from .prepare_bulk_frames import prepare_bulk_frames

if TYPE_CHECKING:
    from endoreg_db.models import (
        VideoPredictionMeta,
        Video,
        RawVideoFile,
        RawVideoPredictionMeta,
        LabelRawVideoSegment,
        LabelVideoSegment,
        ModelMeta,
        Center,
        EndoscopyProcessor,
        VideoMeta,
        Frame,
        RawFrame,
        PatientExamination,
    )  #
    from django.db.models import QuerySet

TEST_RUN = os.environ.get("TEST_RUN", "False")
TEST_RUN = TEST_RUN.lower() == "true"

TEST_RUN_FRAME_NUMBER = int(os.environ.get("TEST_RUN_FRAME_NUMBER", "500"))

if TEST_RUN:
    ic("-----\nTEST RUN ENABLED\n-----")


class AbstractVideoFile(models.Model):
    """
    Abstract base class for video files.
    """

    uuid = models.UUIDField()

    sensitive_meta = models.OneToOneField(
        "SensitiveMeta", on_delete=models.CASCADE, blank=True, null=True
    )

    center = models.ForeignKey("Center", on_delete=models.CASCADE)
    processor = models.ForeignKey(  # TODO Migrate to VideoMeta
        "EndoscopyProcessor", on_delete=models.CASCADE, blank=True, null=True
    )
    # TODO Reduce redundancies between VideoFile and VideoMeta (e.g. center, processor)

    video_meta = models.OneToOneField(
        "VideoMeta", on_delete=models.CASCADE, blank=True, null=True
    )
    examination = models.ForeignKey(
        "PatientExamination", on_delete=models.SET_NULL, blank=True, null=True
    )
    original_file_name = models.CharField(max_length=255)
    video_hash = models.CharField(max_length=255, unique=True)
    uploaded_at = models.DateTimeField(auto_now_add=True)

    frame_dir = models.CharField(max_length=255)
    prediction_dir = models.CharField(max_length=255)
    predictions = models.JSONField(default=dict)
    fps = models.FloatField(blank=True, null=True)
    duration = models.FloatField(blank=True, null=True)
    frame_count = models.IntegerField(blank=True, null=True)

    readable_predictions = models.JSONField(default=dict)
    merged_predictions = models.JSONField(default=dict)
    smooth_merged_predictions = models.JSONField(default=dict)
    binary_smooth_merged_predictions = models.JSONField(default=dict)
    sequences = models.JSONField(default=dict)
    ai_model_meta = models.ForeignKey(
        "ModelMeta", on_delete=models.CASCADE, blank=True, null=True
    )

    # Frame Extraction States
    state_frames_required = models.BooleanField(default=True)
    state_frames_extracted = models.BooleanField(default=False)

    # Video
    ## Prediction
    state_initial_prediction_required = models.BooleanField(default=True)
    state_initial_prediction_completed = models.BooleanField(default=False)
    state_initial_prediction_import_required = models.BooleanField(default=True)
    state_initial_prediction_import_completed = models.BooleanField(default=False)

    # Dataset complete?
    state_histology_required = models.BooleanField(blank=True, null=True)
    state_histology_available = models.BooleanField(default=False)
    state_follow_up_intervention_required = models.BooleanField(blank=True, null=True)
    state_follow_up_intervention_available = models.BooleanField(default=False)
    state_dataset_complete = models.BooleanField(default=False)

    state_frames_initialized = models.BooleanField(default=False)

    is_raw = models.BooleanField(default=False)

    if TYPE_CHECKING:
        self: Union["RawVideoFile", "Video"]
        label_video_segments: Union[
            "QuerySet[LabelVideoSegment]",
            "QuerySet[LabelRawVideoSegment]",
        ]
        examination: "PatientExamination"
        video_meta: "VideoMeta"
        processor: "EndoscopyProcessor"
        center: "Center"
        ai_model_meta: "ModelMeta"
        sensitive_meta: "SensitiveMeta"
        frames: Union["QuerySet[RawFrame]", "QuerySet[Frame]"]

    class Meta:
        abstract = True  #

    @classmethod
    def transcode_videofile(cls, filepath: Path, transcoded_path: Path):
        """ """
        transcoded_path = transcode_videofile(filepath, transcoded_path)
        return transcoded_path

    @classmethod
    def check_hash_exists(cls, video_hash: str):
        return cls.objects.filter(video_hash=video_hash).exists()

    @classmethod
    def create_from_file(  # TODO Rename to get_or_create_from_file
        cls,
        file_path: Path,
        center_name: str,
        processor_name: str,
        frame_dir_parent: Path = FRAME_DIR,
        video_dir: Path = VIDEO_DIR,
        save: bool = True,
        frame_paths: List[dict] = None,
    ):
        from endoreg_db.models import Center, EndoscopyProcessor  # pylint: disable=import-outside-toplevel

        video_dir.mkdir(parents=True, exist_ok=True)
        ic(f"Creating {cls} from {file_path}")
        original_file_name = file_path.name

        transcoded_file_path = transcode_videofile_if_required(file_path)
        video_hash = get_video_hash(transcoded_file_path)

        vid_with_hash_exists = cls.check_hash_exists(video_hash=video_hash)  # pylint: disable=no-member
        if vid_with_hash_exists:
            existing: Union["RawVideoFile", "Video"] = cls.objects.get(
                video_hash=video_hash
            )
            ic(f"Existing DB entry found: {existing}")

            return existing

        _new_file_name, uuid = get_uuid_filename(file_path)
        ic(f"No existing DB entry found, creating new with UUID {uuid}")

        try:
            relative_path = transcoded_file_path.relative_to(STORAGE_LOCATION)
        except ValueError as e:
            raise Exception(
                f"{transcoded_file_path} is outside STORAGE_LOCATION {STORAGE_LOCATION}"
            ) from e

        video = cls(
            uuid=uuid,
            file=relative_path.as_posix(),
            center=Center.objects.get(name=center_name),
            processor=EndoscopyProcessor.objects.get(name=processor_name),
            original_file_name=original_file_name,
            video_hash=video_hash,
        )
        video.save()
        if frame_paths:
            ic(f"Initializing frames using provided paths ({len(frame_paths)})")
            video.initialize_frames(frame_paths)

        # Save the instance to the database
        video.save()
        ic(f"Saved {video}")

        return video

    def get_video_model(self):
        from endoreg_db.models import RawVideoFile, Video

        if self.is_raw:
            return RawVideoFile
        return Video

    def get_frame_model(self):
        from endoreg_db.models import RawFrame, Frame

        if self.is_raw_video_file():
            return RawFrame

        return Frame

    def get_label_segment_model(self):
        from endoreg_db.models import LabelVideoSegment, LabelRawVideoSegment  # pylint: disable=import-outside-toplevel

        if self.is_raw:
            return LabelRawVideoSegment
        return LabelVideoSegment

    def sequences_to_label_video_segments(
        self,
        video_prediction_meta: Union["VideoPredictionMeta", "RawVideoPredictionMeta"],
    ):
        """
        Convert sequences to label video segments.
        """
        from endoreg_db.models import Label, InformationSource

        label_video_segment_model = self.get_label_segment_model()
        for label, sequences in self.sequences.items():
            label = Label.objects.get(name=label)
            for sequence in sequences:
                start_frame_number = sequence[0]
                end_frame_number = sequence[1]

                label_video_segment = label_video_segment_model.objects.create(
                    video=self,
                    prediction_meta=video_prediction_meta,
                    label=label,
                    start_frame_number=start_frame_number,
                    end_frame_number=end_frame_number,
                )
                label_video_segment.save()

    def get_frame_number(self):
        """
        Get the number of frames in the video.
        """
        frame_model = self.get_frame_model()
        framecount = frame_model.objects.filter(video=self).count()
        return framecount

    def set_frames_extracted(self, value: bool = True):
        self.state_frames_extracted = value
        self.save()

    def get_frames(self):
        """
        Retrieve all frames for this video in the correct order.
        """
        frame_model = self.get_frame_model()
        return frame_model.objects.filter(video=self).order_by("frame_number")

    def get_frame(self, frame_number):
        """
        Retrieve a specific frame for this video.
        """
        frame_model = self.get_frame_model()
        return frame_model.objects.get(video=self, frame_number=frame_number)

    def get_frame_range(self, start_frame_number: int, end_frame_number: int):
        """
        Expects numbers of start and stop frame.
        Returns all frames of this video within the given range in ascending order.
        """
        frame_model = self.get_frame_model()
        return frame_model.objects.filter(
            video=self,
            frame_number__gte=start_frame_number,
            frame_number__lte=end_frame_number,
        ).order_by("frame_number")

    def is_raw_video_file(self):
        if self.__class__.__name__ == "RawVideoFile":
            return True
        return False

    def get_prediction_meta_model(self):
        from endoreg_db.models import VideoPredictionMeta, RawVideoPredictionMeta

        if self.is_raw_video_file():
            return RawVideoPredictionMeta
        return VideoPredictionMeta

    def predict_video(
        self,
        model_meta_name: str,
        dataset_name: str = "inference_dataset",
        model_meta_version: Optional[int] = None,
        smooth_window_size_s: int = 1,
        binarize_threshold: float = 0.5,
        anonymized_frames: bool = True,
        img_suffix: str = ".jpg",
        test_run: bool = TEST_RUN,
        n_test_frames: int = TEST_RUN_FRAME_NUMBER,
    ):
        """
        WARNING: When using with Video Objects "anonymous_frames" should be set to False
        Predict the video file using the given model.
        Frames should be extracted and anonymized frames should be generated before prediction.
        """
        from endoreg_db.models import (
            RawVideoFile,
            Video,
            ModelMeta,
            AiModel,
        )  # pylint: disable=import-outside-toplevel
        from endo_ai.predictor.inference_dataset import InferenceDataset  # pylint: disable=import-outside-toplevel
        from endo_ai.predictor.model_loader import MultiLabelClassificationNet
        from endo_ai.predictor.predict import Classifier
        from endo_ai.predictor.postprocess import (
            concat_pred_dicts,
            make_smooth_preds,
            find_true_pred_sequences,
        )

        if TEST_RUN:
            test_run = True

        datasets = {
            "inference_dataset": InferenceDataset,
        }

        if isinstance(self, RawVideoFile):
            self: "RawVideoFile"
        elif isinstance(self, Video):
            self: "Video"
        else:
            raise Exception("Invalid instance type")

        dataset_model_class = datasets[dataset_name]

        if anonymized_frames and self.is_raw:
            try:
                frame_dir = self.get_anonymized_frame_dir()
            except:  # FIXME
                frame_dir = Path(self.frame_dir)
                assert self.state_frames_extracted, "Frames not extracted"

        else:
            frame_dir = Path(self.frame_dir)
            assert self.state_frames_extracted, "Frames not extracted"

        model_meta = ModelMeta.get_by_name(model_meta_name, model_meta_version)
        model: AiModel = model_meta.model

        ic(f"Model: {model}, Model Meta: {model_meta}")

        model_type = model.model_type
        model_subtype = model.model_subtype

        ic(f"Model type: {model_type}, Model subtype: {model_subtype}")
        ic(f"self: {self}, Type: {type(self)}")

        prediction_meta_model = self.get_prediction_meta_model()

        ic(
            f"Prediction Meta Model: {prediction_meta_model}, Type: {type(prediction_meta_model)}"
        )

        video_prediction_meta, _created = prediction_meta_model.objects.get_or_create(
            video=self, model_meta=model_meta
        )

        video_prediction_meta.save()

        ic(video_prediction_meta)

        paths = self.get_frame_paths()
        ic(f"Found {len(paths)} images in {frame_dir}")

        # frame names in format "frame_{index}.jpg"
        indices = [int(p.stem.split("_")[1]) for p in paths]
        path_index_tuples = list(zip(paths, indices))
        # sort ascending by index
        path_index_tuples.sort(key=lambda x: x[1])
        paths, indices = zip(*path_index_tuples)

        crop_template = self.get_crop_template()

        string_paths = [p.resolve().as_posix() for p in paths]
        crops = [crop_template for _ in paths]

        ic(f"Detected {len(paths)} frames")

        if test_run:  # only use the first 10 frames
            ic(f"Running in test mode, using only the first {n_test_frames} frames")
            paths = paths[:n_test_frames]
            indices = indices[:n_test_frames]
            string_paths = string_paths[:n_test_frames]
            crops = crops[:n_test_frames]

        assert paths, f"No images found in {frame_dir}"

        ds_config = model_meta.get_inference_dataset_config()

        # Create dataset
        ds = dataset_model_class(string_paths, crops, config=ds_config)
        ic(f"Dataset length: {len(ds)}")

        # Get a sample image
        sample = ds[0]
        ic("Shape:", sample.shape)  # e.g., torch.Size([3, 716, 716])

        # unorm = get_unorm(ds_config)

        weights_path = model_meta.weights.path

        ic(f"Model path: {weights_path}")

        # FIXME implement support for different model types
        ai_model_instance = MultiLabelClassificationNet.load_from_checkpoint(
            checkpoint_path=weights_path,
        )

        _ = ai_model_instance.cuda()
        _ = ai_model_instance.eval()
        classifier = Classifier(ai_model_instance, verbose=True)

        ic("Starting inference")
        predictions = classifier.pipe(string_paths, crops)

        ic("Creating Prediction Dict")
        prediction_dict = classifier.get_prediction_dict(predictions, string_paths)
        self.predictions = prediction_dict

        ic("Creating Readable Predictions")
        readable_predictions = [classifier.readable(p) for p in predictions]
        self.readable_predictions = readable_predictions

        ic("Creating Merged Predictions")
        merged_predictions = concat_pred_dicts(readable_predictions)

        fps = self.get_fps()
        ic(
            f"Creating Smooth Merged Predictions; FPS: {fps}, \
                Smooth Window Size: {smooth_window_size_s}"
        )

        smooth_merged_predictions = {}
        for key in merged_predictions.keys():
            smooth_merged_predictions[key] = make_smooth_preds(
                prediction_array=merged_predictions[key],
                window_size_s=smooth_window_size_s,
                fps=fps,
            )

        ic(
            "Creating Binary Smooth Merged Predictions; Binarize Threshold: ",
            binarize_threshold,
        )
        binary_smooth_merged_predictions = {}
        for key in smooth_merged_predictions.keys():  # pylint: disable=consider-using-dict-items
            binary_smooth_merged_predictions[key] = (
                smooth_merged_predictions[key] > binarize_threshold
            )

        ic("Creating Sequences")
        sequences = {}
        for label, prediction_array in binary_smooth_merged_predictions.items():
            sequences[label] = find_true_pred_sequences(prediction_array)

        self.sequences = sequences

        self.sequences_to_label_video_segments(
            video_prediction_meta=video_prediction_meta
        )

        ic("Finished inference")
        ic("Saving predictions to DB")
        ic(sequences)
        self.state_initial_prediction_required = False
        self.state_initial_prediction_completed = True
        self.save()

    def get_outside_segments(self, outside_label_name: str = "outside"):
        """
        Get sequences of outside frames.
        """
        from endoreg_db.models import Label  # pylint: disable=import-outside-toplevel

        outside_label = Label.objects.get(name="outside")
        assert outside_label is not None

        outside_segments = self.label_video_segments.filter(label=outside_label)

        ic(f"Getting outside sequences using label: {outside_label}")

        return outside_segments

    def get_outside_frames(self) -> List["Frame"]:
        """
        Get outside frames.
        """
        outside_segments = self.get_outside_segments()
        frames = []
        for segment in outside_segments:
            frames.extend(segment.get_frames())
        return frames

    def get_outside_frame_paths(self):
        """
        Get paths to outside frames.
        """
        frames = self.get_outside_frames()

        frame_paths = [Path(frame.image.path) for frame in frames]

        return frame_paths

    def __str__(self):
        return self.file.name

    def delete_with_file(self):
        file_path = Path(self.file.path)
        if file_path.exists():
            file_path.unlink()
        self.delete_frames()
        self.delete_frames_anonymized()
        self.delete()
        return f"Deleted {self.file.name}; Deleted frames; Deleted anonymized frames"

    def get_endo_roi(self):
        """
        Fetches the endoscope ROI from the video meta.
        Returns a dictionary with keys "x", "y", "width", "height"
        """
        endo_roi = self.video_meta.get_endo_roi()
        return endo_roi

    def get_crop_template(self):
        """
        Creates a crop template (e.g., [0, 1080, 550, 1920 - 20] for a 1080p frame) from the endoscope ROI.
        """
        endo_roi = self.get_endo_roi()
        x = endo_roi["x"]
        y = endo_roi["y"]
        width = endo_roi["width"]
        height = endo_roi["height"]

        crop_template = [y, y + height, x, x + width]
        return crop_template

    def set_frame_dir(self):
        self.frame_dir = f"{FRAME_DIR}/{self.uuid}"

    # video meta should be created when video file is created
    def save(self, *args, **kwargs):
        assert self.processor is not None, "Processor must be set"
        if not self.fps:
            self.fps = self.get_fps()
        if self.is_raw_video_file():
            self.is_raw = True
        if self.video_meta is None:
            center = self.center
            processor = self.processor
            self.video_meta = VideoMeta.objects.create(
                center=center, processor=processor
            )
            self.video_meta.initialize_ffmpeg_meta(self.file.path)

        if not self.frame_dir:
            self.set_frame_dir()

        sm = self.sensitive_meta
        if sm:
            self.patient = sm.pseudo_patient
            self.patient_examination = sm.pseudo_examination

        super(AbstractVideoFile, self).save(*args, **kwargs)

    def extract_frames(
        self,
        quality: int = 2,
        overwrite: bool = False,
        ext="jpg",
        verbose=False,
    ) -> List[Path]:
        """
        Extract frames from the video file and save them to the frame_dir.
        For this, ffmpeg must be available in in the current environment.
        """

        return extract_frames(
            video=self,
            quality=quality,
            overwrite=overwrite,
            ext=ext,
            verbose=verbose,
        )

    def initialize_frames(self, paths: List[Path]):
        """
        Initialize frame objects for the video file.
        """
        initialize_frame_objects(self, paths)

    def delete_frames(self):
        """
        Delete frames extracted from the video file.
        """
        frame_dir = Path(self.frame_dir)
        if frame_dir.exists():
            shutil.rmtree(frame_dir)
            self.state_frames_extracted = False
            self.save()
            return f"Frames deleted from {frame_dir}"
        else:
            return f"No frames to delete for {self.file.name}"

    def delete_frames_anonymized(self):
        """
        Delete anonymized frames extracted from the video file.
        """
        frame_dir = Path(self.frame_dir)
        anonymized_frame_dir = frame_dir.parent / f"anonymized_{self.uuid}"
        if anonymized_frame_dir.exists():
            shutil.rmtree(anonymized_frame_dir)
            return f"Anonymized frames deleted from {anonymized_frame_dir}"
        else:
            return f"No anonymized frames to delete for {self.file.name}"

    def get_frame_path(self, n: int = 0, anonymized=False):
        """
        Get the path to the n-th frame extracted from the video file.
        Note that the frame numbering starts at 1 in our naming convention.
        """
        # Adjust index
        n = n + 1

        if anonymized:
            _frame_dir = Path(self.frame_dir)
            frame_dir = _frame_dir.parent / f"anonymized_{_frame_dir.name}"
        else:
            frame_dir = Path(self.frame_dir)
        return frame_dir / f"frame_{n:07d}.jpg"

    def get_frame_paths(self):
        """
        Get paths to frames extracted from the video file.
        """
        frames = self.frames.filter(extracted=True).order_by("frame_number")

        paths = [Path(frame.image.path) for frame in frames]
        return paths

    def get_prediction_dir(self):
        return Path(self.prediction_dir)

    def get_predictions_path(self, suffix=".json"):
        pred_dir = self.get_prediction_dir()
        return pred_dir.joinpath("predictions").with_suffix(suffix)

    def get_smooth_predictions_path(self, suffix=".json"):
        pred_dir = self.get_prediction_dir()
        return pred_dir.joinpath("smooth_predictions").with_suffix(suffix)

    def get_binary_predictions_path(self, suffix=".json"):
        pred_dir = self.get_prediction_dir()
        return pred_dir.joinpath("binary_predictions").with_suffix(suffix)

    def get_raw_sequences_path(self, suffix=".json"):
        pred_dir = self.get_prediction_dir()
        return pred_dir.joinpath("raw_sequences").with_suffix(suffix)

    def get_filtered_sequences_path(self, suffix=".json"):
        pred_dir = self.get_prediction_dir()
        return pred_dir.joinpath("filtered_sequences").with_suffix(suffix)

    def extract_text_information(self, frame_fraction: float = 0.001):
        """
        Extract text information from the video file.
        Makes sure that frames are extracted and then processes the frames.
        gets all frames from frame_dir and selects a fraction of them to process (at least 1)
        """
        if not self.state_frames_extracted:
            print(f"Frames not extracted for {self.file.name}")
            return None

        processor = self.processor

        frame_paths = self.get_frame_paths()
        n_frames = len(frame_paths)
        n_frames_to_process = max(1, int(frame_fraction * n_frames))

        # Select evenly spaced frames
        frame_paths = frame_paths[:: n_frames // n_frames_to_process]

        # extract text from each frame and store the value to
        # defaultdict of lists.
        # Then, extract the most frequent value from each list
        # Finally, return the dictionary of most frequent values

        # Create a defaultdict to store the extracted text from each ROI
        rois_texts = defaultdict(list)

        print(f"Processing {n_frames_to_process} frames from {self.file.name}")
        # Process frames
        for frame_path in frame_paths[:n_frames_to_process]:
            extracted_texts = extract_text_from_rois(frame_path, processor)
            for roi, text in extracted_texts.items():
                rois_texts[roi].append(text)

        # Get the most frequent text values for each ROI using Counter
        for key in rois_texts.keys():  # pylint: disable=consider-using-dict-items
            counter = Counter([text for text in rois_texts[key] if text])
            rois_texts[key] = counter.most_common(1)[0][0] if counter else None

        return rois_texts

    def update_text_metadata(self, ocr_frame_fraction=0.001):
        ic(f"Updating metadata for {self.file.name}")
        extracted_data_dict = self.extract_text_information(ocr_frame_fraction)
        if extracted_data_dict is None:
            ic("No text extracted; skipping metadata update.")
            return
        extracted_data_dict["center_name"] = self.center.name

        ic(extracted_data_dict)

        extracted_data_dict["center_name"] = self.center.name

        ic("____________")
        ic(extracted_data_dict)
        ic("____________")

        self.sensitive_meta = SensitiveMeta.create_from_dict(extracted_data_dict)
        self.state_sensitive_data_retrieved = True  # pylint: disable=attribute-defined-outside-init
        self.save()

        # Resulting dict depends on defined ROIs for this processor type!

    def update_video_meta(self):
        video_meta = self.video_meta
        video_path = Path(self.file.path)
        center = self.center
        assert self.processor

        if video_meta is None:
            video_meta = VideoMeta.create_from_file(
                video_path,
                center=center,
                processor=self.processor,
            )
            self.video_meta = video_meta
            self.save()

        else:
            video_meta.update_meta(video_path)

    def get_fps(self):
        # # FIXME
        # fps = 50
        # return fps

        if self.video_meta is None:
            self.update_video_meta()

        if self.video_meta.ffmpeg_meta is None:
            self.video_meta.initialize_ffmpeg_meta(self.file.path)

        return self.video_meta.get_fps()

    def create_frame_object(
        self, frame_number, image_file=None, extracted: bool = False
    ):
        """
        Returns a frame instance with the image_file set.
        """
        frame_model = self.get_frame_model()
        return frame_model(
            video=self,
            frame_number=frame_number,
            image=image_file,
            extracted=extracted,
        )

    def bulk_create_frames(self, frames_to_create):
        """
        Bulk create frames, then save their images to storage.
        """
        frame_model = self.get_frame_model()
        created = frame_model.objects.bulk_create(frames_to_create)
        # for frame in created:
        #     frame_name = f"frame_{frame.frame_number:07d}.jpg"
        #     frame.image.save(frame_name, frame.image)
        #     frame.save()



from django.db import models

DEFAULT_PATIENT_LAB_SAMPLE_TYPE_NAME = "generic"

class PatientLabSampleTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class PatientLabSampleType(models.Model):
    """
    A class representing a patient lab sample type.

    Attributes:
        name (str): The name of the patient lab sample type.
        description (str): A description of the patient lab sample type.

    """
    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, null=True)
    name_en = models.CharField(max_length=255, null=True)
    description = models.TextField(blank=True, null=True)

    objects = PatientLabSampleTypeManager()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name
    
    @classmethod
    def get_default_sample_type(cls):
        """
        Get the default patient lab sample type.

        Returns:
            PatientLabSampleType: The default patient lab sample type.

        """
        return cls.objects.get_or_create(name="default")[0]
    
from datetime import datetime as dt    
from datetime import timezone
class PatientLabSample(models.Model):
    """
    A class representing a patient lab sample.

    Attributes:
        patient (Patient): The patient to which the lab sample belongs.
        sample_type (PatientLabSampleType): The type of the lab sample.
        date (datetime): The date of the lab sample.
        values (PatientLabValue; One2Many): The value of the lab sample.
        unit (Unit): The unit of the lab sample.

    """
    patient = models.ForeignKey("Patient", on_delete=models.CASCADE, related_name="lab_samples")
    sample_type = models.ForeignKey("PatientLabSampleType", on_delete=models.CASCADE)
    date = models.DateTimeField()

    def __str__(self):

        formatted_datetime = self.date.strftime('%Y-%m-%d %H:%M')
        return f"{self.patient} - {self.sample_type} - {formatted_datetime} ()"
    
    def get_values(self):
        return self.values.all()
    
    @classmethod
    def create_by_patient(cls, patient=None, sample_type=None, date=None, save = True):
        """
        Create a new patient lab sample by patient.

        Args:
            patient (Patient): The patient to which the lab sample belongs.
            sample_type (PatientLabSampleType): The type of the lab sample.
            date (datetime): The date of the lab sample.

        Returns:
            PatientLabSample: The new patient lab sample.

        """
        from endoreg_db.models.persons.patient import Patient, PatientLabSampleType
        from warnings import warn

        if not patient:
            warn("No patient given. Cannot create patient lab sample.")
            return None
        if not sample_type:
            sample_type = PatientLabSampleType.get_default_sample_type()
        else:
            sample_type = PatientLabSampleType.objects.get(name=sample_type)
        if not date:
            date = dt.now(timezone.utc)

        patient_lab_sample = cls.objects.create(
            patient=patient,
            sample_type=sample_type,
            date=date
        )

        if save:
            patient_lab_sample.save()

        return patient_lab_sample


    
    



'''Model for the medication.'''
from django.db import models


class MedicationManager(models.Manager):
    '''Manager for the medication model.'''
    def get_by_natural_key(self, name):
        '''Retrieve a medication by its natural key.'''
        return self.get(name=name)

class Medication(models.Model):
    '''Model representing a medication.'''
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    adapt_to_renal_function = models.BooleanField(default = False)
    adapt_to_hepatic_function = models.BooleanField(default=False)
    adapt_to_indication = models.BooleanField(default=False)
    adapt_to_age = models.BooleanField(default=False)
    adapt_to_weight = models.BooleanField(default=False)
    adapt_to_risk = models.BooleanField(default=False)
    default_unit = models.ForeignKey('Unit', on_delete=models.CASCADE)


    objects = MedicationManager()

    def natural_key(self):
        '''Return the natural key for the medication.'''
        return (self.name,)

    def __str__(self):
        return str(self.name)
    



from django.db import models

class CenterResource(models.Model):
    name = models.CharField(max_length=255, null=True)
    center = models.ForeignKey("Center", on_delete=models.CASCADE)
    name_de = models.CharField(max_length=255, null=True)
    name_en = models.CharField(max_length=255, null=True)
    quantity = models.FloatField()
    resource = models.ForeignKey("Resource", on_delete=models.CASCADE)
    transport_emission_factor = models.ForeignKey("EmissionFactor", on_delete=models.SET_NULL, null=True,
                                                  related_name="center_resource_transport_emission_factor")
    use_emission_factor = models.ForeignKey("EmissionFactor", on_delete=models.SET_NULL, null=True,
                                            related_name="center_resource_use_emission_factor")
    year = models.IntegerField()
    unit = models.ForeignKey("Unit", on_delete=models.SET_NULL, null=True)
    
    def __str__(self):
        result_string = ""

        if self.name is not None:
            result_string += self.name + ":\n"

        result_string += "\tCenter\t-\t" + str(self.center) + "\n"
        result_string += "\tResource\t-\t" + str(self.resource) + "\n"
        result_string += "\tQuantity\t-\t" + str(self.quantity) + "\n"
        result_string += "\tYear\t-\t" + str(self.year) + "\n"
        result_string += "\tUnit\t-\t" + str(self.unit) + "\n"
        result_string += "\tTransport Emission Factor\t-\t" + str(self.transport_emission_factor) + "\n"
        result_string += "\tUse Emission Factor\t-\t" + str(self.use_emission_factor) + "\n"

        result_string += "\n"

        return result_string


from django.db import models

class LxPermissionManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
#TODO implement base population for permission

class LxPermission(models.Model):
    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True)
    objects = LxPermissionManager()
    
    def __str__(self):
        return self.name
    
    def natural_key(self):
        return (self.name,)



from .transcode_videofile import (
    transcode_videofile,
    transcode_videofile_if_required,
)

from .extract_frames import extract_frames, initialize_frame_objects

__all__ = [
    "extract_frames",
    "initialize_frame_objects",
    "transcode_videofile",
    "transcode_videofile_if_required",
]



from rest_framework import serializers
from endoreg_db.models import Video, VideoImportMeta, LabelVideoSegment


class VideoSerializer(serializers.ModelSerializer):
    class Meta:
        model = Video
        fields = "__all__"


class VideoImportMetaSerializer(serializers.ModelSerializer):
    class Meta:
        model = VideoImportMeta
        fields = "__all__"


class LabelVideoSegmentSerializer(serializers.ModelSerializer):
    class Meta:
        model = LabelVideoSegment
        fields = "__all__"



from django.db import models
import subprocess
import json
from pathlib import Path
from typing import Optional, TYPE_CHECKING

# import endoreg_center_id from django settings
from django.conf import settings

# check if endoreg_center_id is set
if not hasattr(settings, "ENDOREG_CENTER_ID"):
    ENDOREG_CENTER_ID = 9999
else:
    ENDOREG_CENTER_ID = settings.ENDOREG_CENTER_ID

if TYPE_CHECKING:
    from endoreg_db.models import EndoscopyProcessor, Endoscope, Center


# VideoMeta
class VideoMeta(models.Model):
    processor = models.ForeignKey(
        "EndoscopyProcessor", on_delete=models.CASCADE, blank=True, null=True
    )
    endoscope = models.ForeignKey(
        "Endoscope", on_delete=models.CASCADE, blank=True, null=True
    )
    center = models.ForeignKey("Center", on_delete=models.CASCADE)
    import_meta = models.OneToOneField(
        "VideoImportMeta", on_delete=models.CASCADE, blank=True, null=True
    )
    ffmpeg_meta = models.OneToOneField(
        "FFMpegMeta", on_delete=models.CASCADE, blank=True, null=True
    )  ##

    @classmethod
    def create_from_file(
        cls,
        file_path: Path,
        center: Optional["Center"],
        processor: Optional["EndoscopyProcessor"] = None,
        endoscope: Optional["Endoscope"] = None,
    ):
        """Create a new VideoMeta from a file."""
        meta = cls.objects.create(center=center)
        meta.update_meta(file_path)

        if processor:
            meta.processor = processor
            meta.get_endo_roi()
        if endoscope:
            meta.endoscope = endoscope

        meta.save()

        return meta

    def __str__(self):
        processor_name = self.processor.name if self.processor is not None else "None"
        endoscope_name = self.endoscope.name if self.endoscope is not None else "None"
        center_name = self.center.name if self.center is not None else "None"
        ffmpeg_meta_str = self.ffmpeg_meta.__str__()
        import_meta_str = self.import_meta.__str__()

        result_html = ""

        result_html += f"Processor: {processor_name}\n"
        result_html += f"Endoscope: {endoscope_name}\n"
        result_html += f"Center: {center_name}\n"
        result_html += f"FFMpeg Meta: {ffmpeg_meta_str}\n"
        result_html += f"Import Meta: {import_meta_str}\n"

        return result_html

    # import meta should be created when video meta is created
    def save(self, *args, **kwargs):
        if self.import_meta is None:
            self.import_meta = VideoImportMeta.objects.create()
        super(VideoMeta, self).save(*args, **kwargs)

    def initialize_ffmpeg_meta(self, file_path):
        """Initializes FFMpeg metadata for the video file if not already done."""
        self.ffmpeg_meta = FFMpegMeta.create_from_file(Path(file_path))
        self.save()

    def update_meta(self, file_path):
        """Updates the video metadata from the file."""
        self.initialize_ffmpeg_meta(file_path)
        self.save()

    def get_endo_roi(self):
        from endoreg_db.models import EndoscopyProcessor

        processor: EndoscopyProcessor = self.processor
        endo_roi = processor.get_roi_endoscope_image()
        return endo_roi

    def get_fps(self):
        if not self.ffmpeg_meta:
            return None

        return self.ffmpeg_meta.frame_rate


class FFMpegMeta(models.Model):
    # Existing fields
    duration = models.FloatField(blank=True, null=True)
    width = models.IntegerField(blank=True, null=True)
    height = models.IntegerField(blank=True, null=True)
    frame_rate = models.FloatField(blank=True, null=True)

    # New fields for comprehensive information
    video_codec = models.CharField(max_length=50, blank=True, null=True)
    audio_codec = models.CharField(max_length=50, blank=True, null=True)
    audio_channels = models.IntegerField(blank=True, null=True)
    audio_sample_rate = models.IntegerField(blank=True, null=True)

    # Existing __str__ method can be updated to include new fields

    @classmethod
    def create_from_file(cls, file_path: Path):
        cmd = [
            "ffprobe",
            "-v",
            "quiet",
            "-print_format",
            "json",
            "-show_streams",
            str(file_path),
        ]
        proc = subprocess.run(cmd, capture_output=True, text=True, check=True)
        probe = json.loads(proc.stdout)

        video_stream = next(
            (stream for stream in probe["streams"] if stream["codec_type"] == "video"),
            None,
        )
        audio_streams = [
            stream for stream in probe["streams"] if stream["codec_type"] == "audio"
        ]

        if not video_stream:
            return None

        metadata = {
            "duration": float(video_stream.get("duration", 0)),
            "width": int(video_stream.get("width", 0)),
            "height": int(video_stream.get("height", 0)),
            "frame_rate": float(
                next(iter(video_stream.get("avg_frame_rate", "").split("/")), 0)
            ),
            "video_codec": video_stream.get("codec_name", ""),
        }

        if audio_streams:
            first_audio_stream = audio_streams[0]
            metadata.update(
                {
                    "audio_codec": first_audio_stream.get("codec_name", ""),
                    "audio_channels": int(first_audio_stream.get("channels", 0)),
                    "audio_sample_rate": int(first_audio_stream.get("sample_rate", 0)),
                }
            )

        return cls.objects.create(**metadata)

    def __str__(self):
        result_html = ""

        result_html += f"Duration: {self.duration}\n"
        result_html += f"Width: {self.width}\n"
        result_html += f"Height: {self.height}\n"
        result_html += f"Frame Rate: {self.frame_rate}\n"
        result_html += f"Video Codec: {self.video_codec}\n"
        result_html += f"Audio Codec: {self.audio_codec}\n"
        result_html += f"Audio Channels: {self.audio_channels}\n"
        result_html += f"Audio Sample Rate: {self.audio_sample_rate}\n"

        return result_html


class VideoImportMeta(models.Model):
    video_anonymized = models.BooleanField(default=False)
    video_patient_data_detected = models.BooleanField(default=False)
    outside_detected = models.BooleanField(default=False)
    patient_data_removed = models.BooleanField(default=False)
    outside_removed = models.BooleanField(default=False)

    def __str__(self):
        result_html = ""

        result_html += f"Video anonymized: {self.video_anonymized}\n"
        result_html += (
            f"Video patient data detected: {self.video_patient_data_detected}\n"
        )
        result_html += f"Outside detected: {self.outside_detected}\n"
        result_html += f"Patient data removed: {self.patient_data_removed}\n"
        result_html += f"Outside removed: {self.outside_removed}\n"
        return result_html



from .raw_video import RawVideoFile
from .raw_pdf import RawPdfFile

# FileImporter class
# This class is used to import data from a file into the database.
# Expects a directory containing files to import.
# creates correct import file object depending on file type

# FileImporter class
# This class is used to import data from a file into the database by creating objects for the files.
# main method is import_files which expects a path to a directory containing files to import.
# creates correct import file object depending on file type by checking the file extension


__all__ = [
    "RawPdfFile",
    "RawVideoFile",
]



import shutil
import subprocess
from pathlib import Path
from django.db import models
from typing import TYPE_CHECKING, List, Tuple

from icecream import ic
from tqdm import tqdm
import cv2
from django.core.validators import FileExtensionValidator
from django.core.files.storage import FileSystemStorage

from endoreg_db.utils.validate_endo_roi import validate_endo_roi
from ..base_classes.utils import (
    anonymize_frame,
    RAW_VIDEO_DIR_NAME,
    VIDEO_DIR,
    STORAGE_LOCATION,
)
from ..base_classes.abstract_video import AbstractVideoFile

if TYPE_CHECKING:
    # import Queryset
    from django.db.models import QuerySet
    from endoreg_db.models import (
        SensitiveMeta,
        LabelVideoSegment,
    )


# pylint: disable=attribute-defined-outside-init,no-member
class RawVideoFile(AbstractVideoFile):
    """ """

    file = models.FileField(
        upload_to=RAW_VIDEO_DIR_NAME,
        validators=[FileExtensionValidator(allowed_extensions=["mp4"])],  # FIXME
        storage=FileSystemStorage(location=STORAGE_LOCATION.resolve().as_posix()),
    )

    patient = models.ForeignKey(
        "Patient", on_delete=models.SET_NULL, blank=True, null=True
    )

    sensitive_meta = models.ForeignKey(
        "SensitiveMeta",
        on_delete=models.SET_NULL,
        related_name="raw_videos",
        null=True,
        blank=True,
    )

    video = models.ForeignKey(
        "Video",
        on_delete=models.SET_NULL,
        related_name="raw_videos",
        null=True,
        blank=True,
    )

    if TYPE_CHECKING:
        sensitive_meta: "SensitiveMeta"
        label_video_segments: "QuerySet[LabelVideoSegment]"

    # Crop Frames
    state_anonymized_frames_generated = models.BooleanField(default=False)

    ## OCR
    state_ocr_required = models.BooleanField(default=True)
    state_ocr_completed = models.BooleanField(default=False)
    ## Validation
    state_outside_validated = models.BooleanField(default=False)
    state_ocr_result_validated = models.BooleanField(default=False)

    state_sensitive_data_retrieved = models.BooleanField(default=False)

    # Censor Outside
    state_censor_outside_required = models.BooleanField(default=True)
    state_censor_outside_completed = models.BooleanField(default=False)
    state_make_anonymized_video_required = models.BooleanField(default=True)
    state_make_anonymized_video_completed = models.BooleanField(default=False)

    def get_anonymized_video_path(self):
        video_dir = VIDEO_DIR
        video_suffix = Path(self.file.path).suffix
        video_name = f"{self.uuid}{video_suffix}"
        anonymized_video_name = f"TMP_anonymized_{video_name}"
        anonymized_video_path = video_dir / anonymized_video_name

        return anonymized_video_path

    def censor_outside_frames(self):
        assert self.state_frames_extracted, "Frames not extracted"
        assert self.state_initial_prediction_completed, (
            "Initial prediction not completed"
        )
        assert self.state_sensitive_data_retrieved, "Sensitive data not retrieved"

        ic(
            "WARNING: Outside validation is not yet implemented and automatically set to true in this function"
        )

        self.state_outside_validated = True
        self.save()

        assert self.state_outside_validated, "Outside validation not completed"

        outside_frame_paths = self.get_outside_frame_paths()

        if not outside_frame_paths:
            ic("No outside frames found")

        else:
            ic(f"Found {len(outside_frame_paths)} outside frames")
            # use cv2 to replace all outside frames with completely black frames

            for frame_path in tqdm(outside_frame_paths):
                frame = cv2.imread(frame_path.as_posix())
                frame.fill(0)
                cv2.imwrite(frame_path.as_posix(), frame)

        self.state_censor_outside_required = False
        self.state_censor_outside_completed = True
        self.save()

    def get_anonymized_frame_dir(self):
        anonymized_frame_dir = Path(self.frame_dir).parent / f"tmp_{self.uuid}"
        return anonymized_frame_dir

    def make_temporary_anonymized_frames(self) -> Tuple[Path, List[Path]]:
        anonymized_frame_dir = self.get_anonymized_frame_dir()

        assert self.state_frames_extracted, "Frames not extracted"
        assert self.processor, "Processor not set"

        anonymized_frame_dir.mkdir(parents=True, exist_ok=True)
        endo_roi = self.get_endo_roi()
        assert validate_endo_roi(endo_roi), "Endoscope ROI is not valid"
        generated_frame_paths = []

        all_frames = self.frames.all()
        outside_frames = self.get_outside_frames()  #
        outside_frame_numbers = [frame.frame_number for frame in outside_frames]

        # anonymize frames: copy endo-roi content while making other pixels black. (frames are Path objects to jpgs or pngs)
        for frame in tqdm(all_frames):
            frame_path = Path(frame.image.path)
            frame_name = frame_path.name
            frame_number = frame.frame_number

            if frame_number in outside_frame_numbers:
                all_black = True
            else:
                all_black = False

            target_frame_path = anonymized_frame_dir / frame_name
            anonymize_frame(
                frame_path, target_frame_path, endo_roi, all_black=all_black
            )
            generated_frame_paths.append(target_frame_path)

        return anonymized_frame_dir, generated_frame_paths

    def make_anonymized_video(self):
        """
        Make an anonymized video from the anonymized frames.
        """

        assert self.state_initial_prediction_completed, (
            "Initial prediction not completed"
        )
        assert self.state_sensitive_data_retrieved, "Sensitive data not retrieved"

        ic(
            "WARNING: Outside validation is not yet implemented and automatically set to true in this function"
        )
        self.state_outside_validated = True
        self.save()

        assert self.state_outside_validated, "Outside validation not completed"

        _anonymized_frame_dir, generated_frame_paths = (
            self.make_temporary_anonymized_frames()
        )

        anonymized_video_path = self.get_anonymized_video_path()
        # if anonymized video already exists, delete it
        if anonymized_video_path.exists():
            anonymized_video_path.unlink()

        # Use ffmpeg and the frame paths to create a video
        fps = self.get_fps()
        height, width = cv2.imread(generated_frame_paths[0].as_posix()).shape[:2]
        ic("Assembling anonymized video")
        ic(f"Frame width: {width}, height: {height}")
        ic(f"FPS: {fps}")

        command = [
            "ffmpeg",
            "-y",
            "-pattern_type",
            "glob",
            "-f",
            "image2",
            "-framerate",
            str(fps),
            "-i",
            f"{generated_frame_paths[0].parent.as_posix()}/frame_[0-9]*.jpg",
            "-c:v",
            "libx264",
            "-pix_fmt",
            "yuv420p",
            "-vf",
            f"scale={width}:{height}",
            anonymized_video_path.as_posix(),
        ]

        subprocess.run(command, check=True)
        ic(f"Anonymized video saved at {anonymized_video_path}")

        self.state_make_anonymized_video_required = False
        self.state_make_anonymized_video_completed = True
        self.save()

        return anonymized_video_path, generated_frame_paths

    def delete_frames_anonymized(self):
        """
        Delete anonymized frames extracted from the video file.
        """
        frame_dir = Path(self.frame_dir)
        anonymized_frame_dir = frame_dir.parent / f"anonymized_{self.uuid}"
        if anonymized_frame_dir.exists():
            shutil.rmtree(anonymized_frame_dir)
            return f"Anonymized frames deleted from {anonymized_frame_dir}"
        else:
            return f"No anonymized frames to delete for {self.file.name}"

    def get_or_create_video(self):
        from endoreg_db.models import Video, Patient, PatientExamination

        video = self.video
        expected_path = self.get_anonymized_video_path()
        if not video:
            video_hash = self.video_hash
            if Video.objects.filter(video_hash=video_hash).exists():
                video = Video.objects.filter(video_hash=video_hash).first()

            else:
                if not expected_path.exists():
                    ic(
                        f"No anonymized video found at {expected_path}, Creating new one"
                    )
                    video_path, frame_paths = self.make_anonymized_video()

                else:
                    ic(f"Anonymized video found at {expected_path}")
                    video_path = expected_path
                    frame_dir = self.get_anonymized_frame_dir()
                    ic(f"Frame dir: {frame_dir}")
                    frame_paths = list(frame_dir.glob("*.jpg"))
                    ic(f"Found {len(frame_paths)} frames")

                video_object = Video.create_from_file(
                    video_path,
                    self.center,
                    self.processor,
                    video_dir=VIDEO_DIR,
                    frame_paths=frame_paths,
                )

                ex: PatientExamination = self.sensitive_meta.pseudo_examination
                pat: Patient = self.sensitive_meta.pseudo_patient
                video_object.examination = ex
                video_object.patient = pat

                self.video = video_object
                self.save()
                video_object.sync_from_raw_video()

                ic(f"Video object created: {video_object}")
                return video_object

            self.video = video
            self.save()

        # self.vi
        return video



import pytesseract
# import cv2
from endoreg_db.models import EndoscopyProcessor
import os
from collections import Counter
from tempfile import TemporaryDirectory
import re
from datetime import datetime
from typing import Dict, List
import numpy as np

N_FRAMES_MEAN_OCR = 2

# Helper function to process date strings
def process_date_text(date_text):
    """
    Processes a string of text that represents a date and returns a datetime.date object.

    Args:
        date_text (str): A string of text that represents a date.

    Returns:
        datetime.date: A datetime.date object representing the parsed date, or None if the text cannot be parsed.
    """
    try:
        # Remove any non-digit characters
        date_text_clean = re.sub(r'\D', '', date_text)
        # Reformat to 'ddmmyyyy' if necessary
        if len(date_text_clean) == 8:
            return datetime.strptime(date_text_clean, "%d%m%Y").date()
        elif len(date_text_clean) == 14:
            return datetime.strptime(date_text_clean, "%d%m%Y%H%M%S").date()
    except ValueError:
        # Return None if the text cannot be parsed into a date
        # set date to 1/1/1900
        return datetime.strptime("01011900", "%d%m%Y").date()

# Helper function to process patient names
def process_name_text(name_text):
    """
    Remove all numbers, punctuation, and whitespace from a string of text and return the result.
    """
    name = re.sub(r'[0-9!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~\s]+', '', name_text).strip()
    # capitalize first letter of each word
    name = ' '.join([word.capitalize() for word in name.split()])
    return name


# Helper function to process endoscope type text
def process_general_text(endoscope_text):
    """
    This function takes in a string of text from an endoscope and returns a cleaned version of the text.
    """
    return ' '.join(endoscope_text.split())

def roi_values_valid(roi):
    """
    Check if all values in an ROI dictionary are valid (>=0).
    """
    return all([value >= 0 for value in roi.values()])

# Function to extract text from ROIs
def extract_text_from_rois(image_path, processor:EndoscopyProcessor):
    """
    Extracts text from regions of interest (ROIs) in an image using OCR.

    Args:
        image_path (str): The path to the image file.
        processor (EndoscopyProcessor): An instance of the EndoscopyProcessor class.

    Returns:
        dict: A dictionary containing the extracted text for each ROI.
    """
    # Read the image using OpenCV
    image = cv2.imread(image_path)

    # Initialize the dictionary to hold the extracted text
    extracted_texts = {}

    # Define your ROIs and their corresponding post-processing functions in tuples
    rois_with_postprocessing = [
        ('examination_date', processor.get_roi_examination_date, process_date_text),
        ("patient_first_name", processor.get_roi_patient_first_name, process_name_text),
        ('patient_last_name', processor.get_roi_patient_last_name, process_name_text),
        ('patient_dob', processor.get_roi_patient_dob, process_date_text),
        ('endoscope_type', processor.get_roi_endoscope_type, process_general_text),
        ('endoscope_sn', processor.get_roi_endoscopy_sn, process_general_text),
    ]

    # Extract and post-process text for each ROI
    for roi_name, roi_function, post_process in rois_with_postprocessing:
        # Get the ROI dictionary
        roi = roi_function()
        
        # Check if the ROI has values
        
        if roi_values_valid(roi):
            # Crop the image to the ROI
            x, y, w, h = roi['x'], roi['y'], roi['width'], roi['height']
            roi_cropped = image[y:y+h, x:x+w]
            # Convert to grayscale
            gray = cv2.cvtColor(roi_cropped, cv2.COLOR_BGR2GRAY)

            # Invert colors for white text on black background
            gray = cv2.bitwise_not(gray)

            # Binarize the image - using Otsu's method
            _, binary = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)

            # Dilate the image to improve the contour of the pixelated text
            kernel = np.ones((2,2), np.uint8)
            dilation = cv2.dilate(binary, kernel, iterations=1)

            # OCR configuration: Recognize white text on black background without corrections
            config = '--psm 10 -c tessedit_char_whitelist=0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-üöäÜÖÄß'

            # Use pytesseract to do OCR on the preprocessed ROI
            text = pytesseract.image_to_string(dilation, config=config).strip()

            # Post-process extracted text
            processed_text = post_process(text)
            # processed_text = text
            
            # Store the processed text in the dictionary
            extracted_texts[roi_name] = processed_text

        else:
            pass
            # ic(roi_name)
            # ic(roi)
            # ic("No values for this ROI")

    return extracted_texts


def get_most_frequent_values(rois_texts: Dict[str, List[str]]) -> Dict[str, str]:
    """
    Given a dictionary of ROIs and their corresponding texts, returns a dictionary of the most frequent text for each ROI.

    Args:
        rois_texts: A dictionary where the keys are the names of the ROIs and the values are lists of texts.

    Returns:
        A dictionary where the keys are the names of the ROIs and the values are the most frequent text for each ROI.
    """
    most_frequent = {}
    for key in rois_texts.keys():
        counter = Counter([text for text in rois_texts[key] if text])
        # ic(key)
        # ic(counter)
        most_frequent[key], _ = counter.most_common(1)[0] if counter else (None, None)
    return most_frequent

def process_video(video_path, processor):
    """
    Processes a video file by extracting text from regions of interest (ROIs) in each frame.

    Args:
        video_path (str): The path to the video file to process.
        processor (OCRProcessor): An instance of the OCRProcessor class that defines the ROIs to extract text from.

    Returns:
        dict: A dictionary containing the most frequent text values extracted from each ROI.
    """
    # Create a temporary directory to store frames
    with TemporaryDirectory() as temp_dir:
        # ic(temp_dir)
        # Capture the video
        video = cv2.VideoCapture(video_path)
        success, frame_number = True, 0
        rois_texts = {roi_name: [] for roi_name in processor.get_rois().keys()}
        frames_for_mean_extraction = 0

        while success:
            success, frame = video.read()

            # Check if this is the 200th frame
            if frame_number % 1000 == 0 and success:
                frame_path = os.path.join(temp_dir, f"frame_{frame_number}.jpg")
                cv2.imwrite(frame_path, frame)  # Save the frame as a JPEG file
                # cv2.imwrite(f"_tmp/frame_{frame_number}.jpg", frame)
                
                # Extract text from ROIs
                extracted_texts = extract_text_from_rois(frame_path, processor)
                # ic(extracted_texts)
                
                # Store the extracted text from each ROI
                for key, text in extracted_texts.items():
                    rois_texts[key].append(text)
                frames_for_mean_extraction += 1

            frame_number += 1

            if frames_for_mean_extraction >= N_FRAMES_MEAN_OCR: break

        # Release the video capture object
        video.release()

        # Get the most frequent values for each ROI
        return get_most_frequent_values(rois_texts)




from django.db import models


class CenterManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class Center(models.Model):
    objects = CenterManager()

    # import_id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)

    first_names = models.ManyToManyField(
        "FirstName",
        related_name="centers",
    )
    last_names = models.ManyToManyField("LastName", related_name="centers")

    @classmethod
    def get_by_name(cls, name):
        return cls.objects.get(name=name)

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name

    def get_first_names(self):
        from endoreg_db.models import FirstName

        names = FirstName.objects.filter(centers=self)
        return names

    def get_last_names(self):
        from endoreg_db.models import LastName

        names = LastName.objects.filter(centers=self)
        return names

    def get_endoscopes(self):
        from endoreg_db.models import Endoscope

        endoscopes = Endoscope.objects.filter(center=self)
        return endoscopes

    # def get_endoscopy_processor



'''Model for medication intake time'''
from django.db import models


class MedicationIntakeTimeManager(models.Manager):
    '''Manager for the medication intake time model.'''
    def get_by_natural_key(self, name):
        '''Retrieve a medication intake time by its natural key.'''
        return self.get(name=name)
    
class MedicationIntakeTime(models.Model):
    '''Model representing a medication intake time.'''
    name = models.CharField(max_length=255)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    repeats = models.CharField(max_length=20, default = "daily")
    time = models.TimeField()

    objects = MedicationIntakeTimeManager()

    def natural_key(self):
        '''Return the natural key for the medication intake time.'''
        return (self.name,)
    
    def __str__(self):
        return self.name + " at " + str(self.time) + " (" + self.repeats + ")"



from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status
from django.http import FileResponse, Http404
from ..models import RawVideoFile, Label
from ..serializers.video_segmentation import VideoFileSerializer,VideoListSerializer,LabelSerializer
import mimetypes
import os


class VideoView(APIView):
    """
    API endpoint to:
    - Fetch video metadata (JSON)
    - Serve the actual video file dynamically
    """

    def get(self, request, video_id=None):
        """
        Handles GET requests:
        - If no `video_id` is provided, return a list of all videos for frontend dropdown.
        - If `Accept: application/json` is in the request headers, return metadata for a specific video.
        - Otherwise, return the video file.
        """
        if video_id is None:
            return self.get_all_videos()
        return self.get_video_details(request, video_id)

    def get_all_videos(self):
        """
        Returns a list of all available videos along with available labels.
        Used to populate the video selection dropdown in Vue.js.
        """
        videos = RawVideoFile.objects.all()
        labels = Label.objects.all()  # Fetch all labels

        video_serializer = VideoListSerializer(videos, many=True)
        label_serializer = LabelSerializer(labels, many=True)  # Serialize labels

        return Response({
            "videos": video_serializer.data,  # List of videos
            "labels": label_serializer.data  # List of labels
        }, status=status.HTTP_200_OK)

    def get_video_details(self, request, video_id):
        """
        Returns metadata for a specific video if `Accept: application/json` is set.
        Otherwise, streams the video file.
        """
        try:
            video_entry = RawVideoFile.objects.get(id=video_id)
            serializer = VideoFileSerializer(video_entry, context={'request': request})

            accept_header = request.headers.get('Accept', '')

            if "application/json" in accept_header:
                return Response(serializer.data, status=status.HTTP_200_OK)

            return self.serve_video_file(video_entry)

        except RawVideoFile.DoesNotExist:
            return Response({"error": "Video not found in the database."}, status=status.HTTP_404_NOT_FOUND)
        except Exception as e:
            return Response({"error": f"Internal error: {str(e)}"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def serve_video_file(self, video_entry):
        """
        Serves the video file dynamically.
        """
        try:
            full_video_path = video_entry.file.path  # Get the correct file path

            if not os.path.exists(full_video_path):
                raise Http404("Video file not found.")

            mime_type, _ = mimetypes.guess_type(full_video_path)  # Determine the content type (e.g., video/mp4, video/avi)
            response = FileResponse(open(full_video_path, "rb"), content_type=mime_type or "video/mp4")

            # Enable video streaming and CORS
            response["Access-Control-Allow-Origin"] = "*"  # Allow frontend access
            response["Access-Control-Allow-Credentials"] = "true"
            response["Accept-Ranges"] = "bytes"  # Enable seeking in video player
            response["Content-Disposition"] = f'inline; filename="{os.path.basename(full_video_path)}"'  # Instructs the browser to play the video instead of downloading it.

            return response

        except Exception as e:
            return Response({"error": f"Internal error: {str(e)}"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)



from ..serializers.video_segmentation import VideoFileSerializer

class VideoLabelView(APIView):
    """
    API to fetch time segments (start & end times in seconds) for a specific label.
    """

    def get(self, request, video_id, label_name):
        """
        Handles GET request to return:
        - Time segments for the selected label.
        - Frame-wise predictions within those segments.
        """
        try:
            video_entry = RawVideoFile.objects.get(id=video_id)
            
            serializer = VideoFileSerializer(video_entry, context={'request': request})
            label_data = serializer.get_label_time_segments(video_entry)

            # Ensure the requested label exists
            if label_name not in label_data:
                return Response({"error": f"Label '{label_name}' not found"}, status=status.HTTP_404_NOT_FOUND)

            return Response({
                "label": label_name,
                "time_segments": label_data[label_name]["time_ranges"],
                "frame_predictions": label_data[label_name]["frame_predictions"]
            }, status=status.HTTP_200_OK)

        except RawVideoFile.DoesNotExist:
            return Response({"error": "Video not found"}, status=status.HTTP_404_NOT_FOUND)

        except Exception as e:
            return Response({"error": f"Internal error: {str(e)}"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


from ..serializers.video_segmentation import LabelSegmentUpdateSerializer, LabelSegmentSerializer

class UpdateLabelSegmentsView(APIView):
    """
    API to update or create label segments for a video.
    """

    def put(self, request, video_id, label_id):
        """
        Handles PUT request to update or create label segments.
        """
        serializer = LabelSegmentUpdateSerializer(data=request.data)

        if serializer.is_valid():
            result = serializer.save()
            return Response({
                "message": "Segments updated successfully",
                "updated_segments": result["updated_segments"],
                "new_segments": result["new_segments"]
            }, status=status.HTTP_200_OK)
        
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)



# Management command to load LuxNix Model Data

from django.core.management.base import BaseCommand

from endoreg_db.models import (
    LxClientType, LxClientTag, LxPermission
)

from ...utils import load_model_data_from_yaml

from ...data import (
    LX_CLIENT_TYPE_DATA_DIR,
    LX_CLIENT_TAG_DATA_DIR,
    LX_PERMISSION_DATA_DIR
)

IMPORT_MODELS = [
    LxPermission.__name__,
    LxClientType.__name__,
    LxClientTag.__name__, # M2M relationship with LxPermission ("permissions")
]

IMPORT_METADATA = {
    LxPermission.__name__: {
        "dir": LX_PERMISSION_DATA_DIR,
        "model": LxPermission,
        "foreign_keys": [],
        "foreign_key_models": []
    },
    LxClientType.__name__: {
        "dir": LX_CLIENT_TYPE_DATA_DIR,
        "model": LxClientType,
        "foreign_keys": [],
        "foreign_key_models": []
    },
    LxClientTag.__name__: {
        "dir": LX_CLIENT_TAG_DATA_DIR,
        "model": LxClientTag,
        "foreign_keys": ["permissions"],
        "foreign_key_models": [LxPermission]
    },
}

class Command(BaseCommand):
    help = """Load all .yaml files in the data/intervention directory
    into the Intervention and InterventionType model"""

    def add_arguments(self, parser):
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='Display verbose output',
        )

    def handle(self, *args, **options):
        verbose = options['verbose']
        for model_name in IMPORT_MODELS:
            _metadata = IMPORT_METADATA[model_name]
            load_model_data_from_yaml(
                self,
                model_name,
                _metadata,
                verbose
            )



from django.db import models

class RulesetManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class Ruleset(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    rules = models.ManyToManyField('Rule')

    objects = RulesetManager()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name



from typing import Union, TYPE_CHECKING
from django.db import models

from endoreg_db.models.label.label import LabelSet
from ..data_file.video import Video
from ..data_file.import_classes import RawVideoFile
from ..data_file.frame import Frame
from .image_classification import ImageClassificationPrediction
from ..data_file import (
    LabelVideoSegment,
    find_segments_in_prediction_array,
)

# from ..information_source import get_prediction_information_source
import numpy as np
import pickle

DEFAULT_WINDOW_SIZE_IN_SECONDS_FOR_RUNNING_MEAN = 1.5
DEFAULT_VIDEO_SEGMENT_LENGTH_THRESHOLD_IN_S = 1.0

if TYPE_CHECKING:
    from endoreg_db.models import ModelMeta


class AbstractVideoPredictionMeta(models.Model):
    model_meta = models.ForeignKey("ModelMeta", on_delete=models.CASCADE)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)
    video = None  # Placeholder for the video field, to be defined in derived classes
    prediction_array = models.BinaryField(blank=True, null=True)
    is_raw = models.BooleanField(default=False)

    if TYPE_CHECKING:
        model_meta: "ModelMeta"
        video: Union["Video", "RawVideoFile"]

    class Meta:
        abstract = True
        unique_together = ("model_meta", "video")

    def is_raw_video_prediction_meta(self):
        if isinstance(self, RawVideoPredictionMeta):
            return True
        return False

    def __str__(self):
        return f"Video {self.video.id} - {self.model_meta.name}"

    # override save method to set is_raw field
    def save(self, *args, **kwargs):
        self.is_raw = self.is_raw_video_prediction_meta()
        super().save(*args, **kwargs)

    def get_labelset(self):
        """
        Get the labelset of the predictions.
        """
        if self.is_raw:
            return self.model_meta.labelset

    def get_video_model(self):
        if self.is_raw:
            return RawVideoFile

        return Video

    def get_frame_model(self):
        return Frame

    def get_video_segment_model(self):
        from endoreg_db.models import LabelRawVideoSegment

        if self.is_raw:
            return LabelRawVideoSegment
        return LabelVideoSegment

    def get_label_list(self):
        """
        Get the label list of the predictions.
        """
        labelset: LabelSet = self.get_labelset()
        label_list = labelset.get_labels_in_order()
        return label_list

    def save_prediction_array(self, prediction_array: np.array):
        """
        Save the prediction array to the database.
        """
        self.prediction_array = pickle.dumps(prediction_array)
        self.save()

    def get_prediction_array(self):
        """
        Get the prediction array from the database.
        """
        if self.prediction_array is None:
            return None
        else:
            return pickle.loads(self.prediction_array)

    def calculate_prediction_array(self):
        assert 1 == 2, "This method should be overridden in derived classes"

    def apply_running_mean(self, confidence_array, window_size_in_seconds: int = None):
        """
        Apply a running mean filter to the confidence array for smoothing, assuming a padding
        of 0.5 for the edges.

        Args:
        self: Object that has video and fps attributes, and to which this function belongs.
        confidence_array: A 2D numpy array with dimensions (num_frames),
                        containing confidence scores for each label at each frame.
        window_size_in_seconds: The window size for the running mean in seconds.

        Returns:
        running_mean_array: A 2D numpy array with the same dimensions as confidence_array,
                            containing the smoothed confidence scores.
        """
        video = self.video
        fps = video.fps

        if not window_size_in_seconds:
            window_size_in_seconds = DEFAULT_WINDOW_SIZE_IN_SECONDS_FOR_RUNNING_MEAN

        # Calculate window size in frames, ensuring at least one frame
        window_size_in_frames = int(window_size_in_seconds * fps)
        window_size_in_frames = max(window_size_in_frames, 1)

        # Define the window for the running mean
        window = np.ones(window_size_in_frames) / window_size_in_frames

        # Create running mean array with the same shape as the confidence array
        _running_mean_array = np.zeros(confidence_array.shape)

        # Calculate the padding size
        pad_size = window_size_in_frames // 2

        # Pad the array with 0.5 on both sides
        padded_confidences = np.pad(
            confidence_array,
            (pad_size, pad_size),
            "constant",
            constant_values=(0.5, 0.5),
        )

        # Apply the running mean filter on the padded array
        running_mean = np.convolve(padded_confidences, window, mode="same")

        # Remove the padding from the result to match the original shape
        running_mean = running_mean[pad_size:-pad_size]

        return running_mean

    # FIXME
    # def create_video_segments_for_label(self, segments, label):
    #     """
    #     Creates video segments for the given label and segments.
    #     Segments is a list of tuples (start_frame_number, end_frame_number).
    #     Labels is a Label object.
    #     """
    #     video = self.video
    #     video_segment_model = self.get_video_segment_model()
    #     information_source = get_prediction_information_source()

    #     for segment in segments:
    #         start_frame_number, end_frame_number = segment

    #         video_segment = video_segment_model(
    #             video=video,
    #             prediction_meta=self,
    #             start_frame_number=start_frame_number,
    #             end_frame_number=end_frame_number,
    #             source=information_source,
    #             label=label,
    #         )
    #         video_segment.save()

    def create_video_segments(self, segment_length_threshold_in_s: float = None):
        if not segment_length_threshold_in_s:
            segment_length_threshold_in_s = DEFAULT_VIDEO_SEGMENT_LENGTH_THRESHOLD_IN_S

        video = self.video
        fps = video.fps
        min_frame_length = int(segment_length_threshold_in_s * fps)

        label_list = self.get_label_list()

        # if prediction array doesnt exist, create it
        if self.prediction_array is None:
            self.calculate_prediction_array()

        prediction_array = self.get_prediction_array()

        for i, label in enumerate(label_list):
            # get predictions for this label
            predictions = prediction_array[:, i]
            # find segments of predictions that are longer than the threshold
            # segments is a list of tuples (start_frame_number, end_frame_number)
            segments = find_segments_in_prediction_array(predictions, min_frame_length)

            # create video segments
            self.create_video_segments_for_label(segments, label)


class RawVideoPredictionMeta(AbstractVideoPredictionMeta):
    """
    Model for storing video predictions for a specific model and video.
    """

    video = models.ForeignKey(
        "RawVideoFile", on_delete=models.CASCADE, related_name="video_prediction_meta"
    )

    def calculate_prediction_array(self, window_size_in_seconds: int = None):
        """
        Fetches all predictions for this video, labelset, and model meta.
        """
        video: Video = self.video

        model_meta = self.model_meta
        label_list = self.get_label_list()

        prediction_array = np.zeros((video.get_frame_number, len(label_list)))
        for i, label in enumerate(label_list):
            # fetch all predictions for this label, video, and model meta ordered by ImageClassificationPrediction.frame.frame_number
            predictions = ImageClassificationPrediction.objects.filter(
                label=label, frame__video=video, model_meta=model_meta
            ).order_by("frame__frame_number")
            confidences = np.array(
                [prediction.confidence for prediction in predictions]
            )
            smooth_confidences = self.apply_running_mean(
                confidences, window_size_in_seconds
            )
            # calculate binary predictions
            binary_predictions = smooth_confidences > 0.5
            # add to prediction array
            prediction_array[:, i] = binary_predictions

        # save prediction array
        self.save_prediction_array(prediction_array)


class VideoPredictionMeta(AbstractVideoPredictionMeta):
    """
    Model for storing video predictions for a specific model and video.
    """

    video = models.ForeignKey(
        "Video", on_delete=models.CASCADE, related_name="video_prediction_meta"
    )

    @classmethod
    def from_raw(
        cls, video: "Video", raw_video_prediction_meta: RawVideoPredictionMeta
    ):
        """
        Create a new VideoPrediction from an existing RawVideoPredictionMeta.
        """
        cls_dict = {
            "video": video,
            "model_meta": raw_video_prediction_meta.model_meta,
            "date_created": raw_video_prediction_meta.date_created,
            "date_modified": raw_video_prediction_meta.date_modified,
            "prediction_array": raw_video_prediction_meta.prediction_array,
            "is_raw": False,
        }

        # check if exists
        if cls.objects.filter(
            video=video, model_meta=raw_video_prediction_meta.model_meta
        ).exists():
            return cls.objects.get(
                video=video, model_meta=raw_video_prediction_meta.model_meta
            )
        else:
            return cls.objects.create(**cls_dict)

    def calculate_prediction_array(self, window_size_in_seconds: int = None):
        """
        Fetches all predictions for this video, labelset, and model meta.
        """
        video: Video = self.video

        model_meta = self.model_meta
        label_list = self.get_label_list()

        prediction_array = np.zeros((video.get_frame_number, len(label_list)))
        for i, label in enumerate(label_list):
            # fetch all predictions for this label, video, and model meta ordered by ImageClassificationPrediction.frame.frame_number
            predictions = ImageClassificationPrediction.objects.filter(
                label=label, frame__video=video, model_meta=model_meta
            ).order_by("frame__frame_number")
            confidences = np.array(
                [prediction.confidence for prediction in predictions]
            )
            smooth_confidences = self.apply_running_mean(
                confidences, window_size_in_seconds
            )
            # calculate binary predictions
            binary_predictions = smooth_confidences > 0.5
            # add to prediction array
            prediction_array[:, i] = binary_predictions

        # save prediction array
        self.save_prediction_array(prediction_array)



from endoreg_db.models import ReportFile
from typing import Optional

def get_report_files() -> ReportFile:
    """
    Returns all ReportFile objects from the database.
    """
    return ReportFile.objects.all()

def get_report_file_by_id(id) -> Optional[ReportFile]:
    """Retrieve a ReportFile object by its id.

    Args:
        id (int): The id of the report_file to retrieve.

    Returns:
        Optional[ReportFile]: The ReportFile object with the given id, or None if it does not exist.
    """
    return ReportFile.objects.get(id=id)

def get_report_files_without_examination() -> ReportFile:
    """
    Returns all ReportFile objects from the database without an examination.
    """
    return ReportFile.objects.filter(patient_examination__isnull=True)

def get_report_files_without_patient() -> ReportFile:
    """
    Returns all ReportFile objects from the database without a patient.
    """
    return ReportFile.objects.filter(patient__isnull=True)





from django.db import models

class LxIdentityTypeManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
#TODO implement base population for identity Type
## ed25519, rsa, ecdsa, dsa, etc

class LxIdentityType(models.Model):
    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True)
    objects = LxIdentityTypeManager()
    
    def __str__(self):
        return self.name
    
    def natural_key(self):
        return (self.name,)

class LxIdentity(models.Model):
    name = models.CharField(max_length=100, unique=True)
    description = models.TextField(blank=True)
    identity_type = models.ForeignKey(LxIdentityType, on_delete=models.CASCADE)
    objects = models.Manager()

    # Foreign key to user, nullable
    user = models.ForeignKey('LxUser', on_delete=models.CASCADE, null=True, blank=True)
    
    def __str__(self):
        return self.name
    
    def natural_key(self):
        return (self.name,)



from django.urls import path
from django.contrib import admin
from django.http import JsonResponse
from endoreg_db.models import (
    Patient,
    Examination,
    PatientExamination,
    Finding,
    FindingLocationClassification,
    FindingMorphologyClassification,
    FindingIntervention,  #  Import Finding Interventions
    PatientFindingIntervention,
    FindingLocationClassificationChoice,
    FindingMorphologyClassificationChoice,
)
from endoreg_db.forms.patient_finding_intervention_form import (
    PatientFindingInterventionForm,
)
from endoreg_db.forms.patient_form import PatientForm


@admin.register(Patient)
class PatientAdmin(admin.ModelAdmin):
    form = PatientForm
    list_display = ("id", "first_name", "last_name", "dob", "center")
    search_fields = ("first_name", "last_name", "email", "phone")
    list_filter = ("dob", "center")
    ordering = ("last_name",)


@admin.register(Examination)
class ExaminationAdmin(admin.ModelAdmin):
    list_display = ("id", "name")
    search_fields = ("name", "name_de", "name_en")
    list_filter = ("name",)
    ordering = ("name",)


class PatientFindingInterventionAdmin(admin.ModelAdmin):
    change_list_template = "admin/patient_finding_intervention.html"

    def changelist_view(self, request, extra_context=None):
        """Pass Patients, Examinations, Findings, Locations, Location Choices, Morphologies, Morphology Choices, and Finding Interventions to the template"""
        extra_context = {
            "patients": Patient.objects.all(),
            "examinations": Examination.objects.all(),
            "findings": Finding.objects.all(),
            "locations": FindingLocationClassification.objects.all(),
            "location_choices": FindingLocationClassificationChoice.objects.none(),
            "morphologies": FindingMorphologyClassification.objects.all(),
            "morphology_choices": FindingMorphologyClassificationChoice.objects.all(),
            "finding_interventions": FindingIntervention.objects.all(),  #  Pass Finding Interventions
        }
        return super().changelist_view(request, extra_context=extra_context)

    def get_location_choices_json(self, request):
        """Ensure Django Admin returns JSON response for Location Choices"""
        location_id = request.GET.get("location")
        if not location_id:
            return JsonResponse({"error": "Location ID is required"}, status=400)

        try:
            choices = list(
                FindingLocationClassificationChoice.objects.filter(
                    location_classifications__id=location_id
                ).values("id", "name")
            )  #  Convert QuerySet to List

            if not choices:
                return JsonResponse(
                    [], safe=False
                )  #  Return an empty list if no data found

            return JsonResponse(choices, safe=False)  #  Return valid JSON

        except Exception as e:
            return JsonResponse({"error": str(e)}, status=500)

    def get_urls(self):
        """Register JSON endpoint inside Django Admin"""
        urls = super().get_urls()
        custom_urls = [
            path(
                "ajax/get-location-choices/",
                self.admin_site.admin_view(self.get_location_choices_json),
                name="ajax_get_location_choices",
            ),
        ]
        return custom_urls + urls


admin.site.register(PatientFindingIntervention, PatientFindingInterventionAdmin)



# Django db class to store network device types (e.g., server, client, switch, etc.)

from django.db import models

class NetworkDeviceTypeManager(models.Manager):
    # Custom manager for NetworkDeviceType; defines name as natural key
    def get_by_natural_key(self, name):
        return self.get(name=name)

class NetworkDeviceType(models.Model):
    name = models.CharField(max_length=255)
    description = models.CharField(max_length=255)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    objects = NetworkDeviceTypeManager()

    def __str__(self):
        return self.name

    class Meta:
        db_table = 'network_device_types'
        ordering = ['name']




from .sensitive_meta import SensitiveMeta
from .pdf_meta import PdfMeta, PdfType
from .video_meta import VideoMeta, FFMpegMeta, VideoImportMeta


__all__ = [
    "SensitiveMeta",
    "PdfMeta",
    "PdfType",
    "VideoMeta",
    "FFMpegMeta",
    "VideoImportMeta",
]



from django.db import models
from typing import List

class PatientFinding(models.Model):
    patient_examination = models.ForeignKey('PatientExamination', on_delete=models.CASCADE, related_name='patient_findings')
    finding = models.ForeignKey('Finding', on_delete=models.CASCADE, related_name='patient_findings')
    locations = models.ManyToManyField(
        'PatientFindingLocation',
        blank=True,
        related_name='patient_findings'
    )
    morphologies = models.ManyToManyField(
        'PatientFindingMorphology',
        blank=True,
        related_name='patient_findings'
    )
    

    class Meta:
        verbose_name = 'Patient Finding'
        verbose_name_plural = 'Patient Findings'
        ordering = ['patient_examination', 'finding']

    def __str__(self):
        return f"{self.patient_examination} - {self.finding}"

   # functions to get all associated location and morphology choices
    def get_locations(self):
        """
        Returns all location choices that are associated with this patient finding.
        """
        from endoreg_db.models import PatientFindingLocation
        locations:List[PatientFindingLocation] = [_ for _ in self.locations.all()]
        return locations
    
    def get_morphologies(self):
        """
        Returns all morphology choices that are associated with this patient finding.
        """
        from endoreg_db.models import PatientFindingMorphology
        morphologies:List[PatientFindingMorphology] = [_ for _ in self.morphologies.all()]
        return morphologies   

    def add_morphology_choice(self, morphology_choice, morphology_classification):
        """
        Adds a morphology choice to this patient finding morphology.
        """
        from endoreg_db.models import (
            FindingMorphologyClassificationChoice,
            FindingMorphologyClassification,
            PatientFindingMorphology
        )
        morphology_choice: FindingMorphologyClassificationChoice
        morphology_classification: FindingMorphologyClassification
        
        patient_finding_morphology = PatientFindingMorphology.objects.create(
            morphology_classification=morphology_classification,
            morphology_choice=morphology_choice
        )
        patient_finding_morphology.save()

        self.morphologies.add(patient_finding_morphology)
        self.save()

        return patient_finding_morphology

    def add_intervention(self, intervention, state="pending", date=None):
        """
        Adds an intervention to this patient finding.
        """
        from endoreg_db.models import PatientFindingIntervention
        patient_finding_intervention = PatientFindingIntervention.objects.create(
            patient_finding=self,
            intervention=intervention,
            state = state,
            date = date
        )
        patient_finding_intervention.save()

        return patient_finding_intervention

    def add_location_choice(self, location_choice, location_classification):
        """
        Adds a location choice to this patient finding location.
        """
        from endoreg_db.models import (
            FindingLocationClassificationChoice,
            FindingLocationClassification,
            PatientFindingLocation
        )
        location_choice: FindingLocationClassificationChoice
        location_classification: FindingLocationClassification
        
        patient_finding_location = PatientFindingLocation.objects.create(
            location_classification=location_classification,
            location_choice=location_choice
        )
        patient_finding_location.save()

        self.locations.add(patient_finding_location)
        self.save()

        return patient_finding_location

    def get_interventions(self):
        """
        Returns all interventions that are associated with this patient finding.
        """
        from endoreg_db.models import PatientFindingIntervention
        interventions:List[PatientFindingIntervention] = [_ for _ in self.interventions.all()]
        return interventions

    def set_random_location(
            self, location_classification
        ):
        """
        Sets a random location for this finding based on the location classification.
        """
        from endoreg_db.models import (
            FindingLocationClassificationChoice,
            FindingLocationClassification, 
            PatientFindingLocation
        )
        import random
        from typing import List
        location_classification:FindingLocationClassification

        # assert location_classification in self.finding.location_classifications.all()

        location_choices:List[FindingLocationClassificationChoice] = location_classification.choices.all()
        location_choice = random.choice(location_choices)

        patient_finding_location = PatientFindingLocation.objects.create(
            location_classification=location_classification,
            location_choice=location_choice
        )

        self.locations.add(patient_finding_location)
        self.save()

        return patient_finding_location
    
    


from django.db import models
import pandas as pd

class ProductWeightManager(models.Manager):
    def get_by_natural_key(self, product, product_group, weight):
        return self.get(product=product, product_group=product_group, weight=weight)
    
class ProductWeight(models.Model):
    objects = ProductWeightManager()

    name = models.CharField(max_length=255, null = True)
    product = models.ForeignKey("Product", on_delete=models.CASCADE)
    measured = models.FloatField(null=True)
    verified = models.FloatField(null=True)
    manufacturer = models.FloatField(null=True)
    unit = models.ForeignKey("Unit", on_delete=models.SET_NULL, null=True)

    def get_weight(self):
        if not pd.isnull(self.verified):
            return self.verified
        elif not pd.isnull(self.measured):
            return self.measured
        elif not pd.isnull(self.manufacturer):
            return self.manufacturer
        else:
            return None
        
    def get_weight_source(self):
        if not pd.isnull(self.verified):
            return "verified"
        elif not pd.isnull(self.measured):
            return "measured"
        elif not pd.isnull(self.manufacturer):
            return "manufacturer"
        else:
            return None
    def __str__(self):
        return f"{self.product} - {self.get_weight()} {self.unit} (Source: {self.get_weight_source()})"



from io import StringIO
from django.core.management import call_command
from django.test import TestCase

from endoreg_db.models import (
    CaseTemplate
)

from endoreg_db.case_generator.case_generator import CaseGenerator

DEFAULT_TEMPLATE_NAME = "pre_default_screening_colonoscopy"

RULE_ATTRIBUTES = [
    "name",
    "rule_type",
    # "parent_model",
    # "parent_field",
    # "target_model",
    # "target_field",
    # "rule_values",
    # "extra_parameters",
    # "value_type",
    "chained_rules"
]

def print_rule_attributes(rule):
    print("Rule:")
    print(rule)
    print("Attributes:")
    
    for attr in RULE_ATTRIBUTES:
        print(f"{attr}: {getattr(rule, attr)}")

class LoadGplayData(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_center_data", stdout=out)
        call_command("load_distribution_data", stdout=out)
        call_command("load_lab_value_data", stdout=out)
        call_command("load_g_play_data", stdout=out)

    def test_check_templates(self):
        templates = CaseTemplate.objects.all()
        assert templates
        # print("Case Templates:")
        # for template in templates:
        #     print(template)

    def test_check_default_template(self):
        template = CaseTemplate.objects.get(name=DEFAULT_TEMPLATE_NAME)
        assert template
        # print("Default Template:")
        # print(template)

        # print("Patient Create Rule:")
        patient_create_rule = template.get_create_patient_rule()
        assert patient_create_rule
        # print_rule_attributes(patient_create_rule)



        cg = CaseGenerator(template)

        # print("Testing create patient rule:")
        patient = cg.apply_rule(patient_create_rule)
        # print(patient)


        


from endoreg_db.models import Center
from typing import Optional

def get_centers() -> Center:
    """
    Returns all Center objects from the database.
    """
    return Center.objects.all()

def get_center_by_name(name) -> Optional[Center]:
    """Retrieve a Center object by its name.

    Args:
        name (str): The name of the center to retrieve.

    Returns:
        Optional[Center]: The Center object with the given name, or None if it does not exist.
    """
    return Center.objects.get(name=name)

def get_center_by_id(id) -> Optional[Center]:
    """Retrieve a Center object by its id.

    Args:
        id (int): The id of the center to retrieve.

    Returns:
        Optional[Center]: The Center object with the given id, or None if it does not exist.
    """
    return Center.objects.get(id=id)

def get_center_by_natural_key(name: str) -> Optional[Center]:
    """
    Retrieve a Center object by its natural key.

    Args:
        name: The name of the center to retrieve.

    Returns:
        The Center object with the given name, or None if it does not exist.
    """
    return Center.objects.get_by_natural_key(name=name)



import shutil
import subprocess
import os
from pathlib import Path
from icecream import ic


def get_transcoded_file_path(source_file_path: Path, suffix: str = "mp4"):
    """
    Method to get the transcoded file path.

    Args:
        source_file_path (Path): Source file path.
        suffix (str): Suffix of the transcoded file.

    Returns:
        transcoded_file_path (Path): Transcoded file path.
    """
    transcoded_file_name = f"{source_file_path.stem}_transcoded.{suffix}"
    transcoded_file_path = source_file_path.parent / transcoded_file_name
    return transcoded_file_path


def check_require_transcode(
    filepath: Path, transcoded_file_path: Path, target_suffix=".mp4"
):
    """
    Checks whether a video file requires transcoding.\
    We check if the current suffix of the file path matches the target suffix\
    and if the transcoded file path exists.\
    If the current suffix does not match the target suffix and the transcoded file path does not exist,\
    transcoding is required.
    """
    current_suffix = filepath.suffix

    require_transcode = False
    if not current_suffix == target_suffix and not transcoded_file_path.exists():
        if not transcoded_file_path.exists():
            require_transcode = True
    return require_transcode


def transcode_videofile_if_required(filepath: Path):
    """
    Perform transcoding on a video file if required.
    This method checks whether a transcoded version (with an ".mp4" suffix) of the given
    video file exists or needs to be produced. It first computes the expected transcoded file path,
    then uses a class-specific check (check_require_transcode) to decide if transcoding is necessary.
    If so, it transcodes the video file by calling the class method transcode_videofile and returns
    the path of the transcoded file after ensuring that the resulting file path matches the computed one.
    If transcoding is not required, the original file path is returned.
    Args:
        filepath (Path): The path to the original video file that may require transcoding.
    Returns:
        Path: The path to the transcoded video file if transcoding was performed; otherwise, the original file path.
    """

    transcoded_file_path = get_transcoded_file_path(filepath, suffix=".mp4")
    if check_require_transcode(filepath, transcoded_file_path):
        transcoded_path = transcode_videofile(
            filepath, transcoded_path=transcoded_file_path
        )
        assert transcoded_file_path == transcoded_path
        return transcoded_path
    else:
        return filepath


def transcode_videofile(filepath: Path, transcoded_path: Path):
    """
    Transcodes a video to a compatible MP4 format using ffmpeg.
    If the transcoded file exists, it is returned.

    Parameters
    ----------
    mov_file : str
        The full path to the video file.

    Returns
    -------
    transcoded_path : str
        The full path to the transcoded video file.
    """
    ic("Transcoding video")
    ic(f"Input path: {filepath}")

    # if filepath suffix is .mp4 or .MP4 we dont need to transcode and can copy the file
    if filepath.suffix.lower() in [".mp4"]:
        shutil.copyfile(filepath, transcoded_path)
        return transcoded_path

    ic(f"Transcoded path: {transcoded_path}")
    if os.path.exists(transcoded_path):
        return transcoded_path

    # Run ffmpeg to transcode the video using H264 and AAC
    # TODO Document settings, check if we need to change them
    command = [
        "ffmpeg",
        "-i",
        filepath.resolve().as_posix(),
        "-c:v",
        "libx264",
        "-preset",
        "fast",
        "-c:a",
        "aac",
        transcoded_path,
    ]
    subprocess.run(command, check=True)
    return transcoded_path



from datetime import datetime, date, timedelta
from random import randint
from calendar import monthrange

# TODO replace used random_day_by_year function implementation when
# creating pseudo patients with new function "random date by age_at_date and examination_date"


def random_day_by_age_at_date(age_at_date: int, examination_date: date) -> date:
    """
    Return a random birth day for a patient with the specified age at the specified examination date.
    """
    examination_year = examination_date.year
    latest_birthdate = examination_date.replace(year=examination_year - age_at_date)
    valid_dates = [latest_birthdate - timedelta(days=i) for i in range(365)]

    select = randint(0, len(valid_dates) - 1)
    birth_date = valid_dates[select]

    return birth_date


def random_day_by_year(year: int) -> date:
    """
    Return a random birth day within the specified year.
    """
    month = randint(1, 12)
    day = randint(1, monthrange(year, month)[1])

    return date(year, month, day)


def random_day_by_month_year(month: int, year) -> date:
    """
    Return a random birth day within the specified month and year.
    """

    day = randint(1, monthrange(year, month)[1])
    return date(year, month, day)



from django.db import models
from typing import List

class DiseaseManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class Disease(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)
    subcategories = models.JSONField(default=dict)
    numerical_descriptors = models.JSONField(default=dict)

    objects = DiseaseManager()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name
    
    

        


    
    def get_classifications(self)->List['DiseaseClassification']:
        classifications: List[DiseaseClassification] = [_ for _ in self.disease_classifications.all()]
        return classifications
    
class DiseaseClassificationManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class DiseaseClassification(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)

    disease = models.ForeignKey(
        Disease, on_delete=models.CASCADE,
        related_name='disease_classifications'
    )

    objects = DiseaseClassificationManager()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name
    
    def get_choices(self)->List['DiseaseClassificationChoice']:
        choices:List[DiseaseClassificationChoice] = [_ for _ in self.disease_classification_choices.all()]
        return choices
    
class DiseaseClassificationChoiceManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)

class DiseaseClassificationChoice(models.Model):
    name = models.CharField(max_length=255, unique=True)
    name_de = models.CharField(max_length=255, blank=True, null=True)
    name_en = models.CharField(max_length=255, blank=True, null=True)

    disease_classification = models.ForeignKey(
        DiseaseClassification, on_delete=models.CASCADE,
        related_name='disease_classification_choices'
    )

    objects = DiseaseClassificationChoiceManager()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name


from django.db import models

class PatientMedicationSchedule(models.Model):
    patient = models.ForeignKey("Patient", on_delete= models.CASCADE)
    medication = models.ManyToManyField(
        'PatientMedication', 
        related_name='patient_medication_schedules',
        blank=True
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return f'{self.patient} - {self.medication.all()}'
    
    @classmethod
    def create_by_patient_and_indication_type(cls, patient, indication_type):
        from endoreg_db.models import MedicationIndicationType, PatientMedication

        medication_indication = MedicationIndicationType.get_random_indication_by_type(name=indication_type)

        patient_medication_schedule = cls.objects.create(patient=patient)
        patient_medication_schedule.save()

        patient_medication = PatientMedication.create_by_patient_and_indication(patient, medication_indication)
        patient_medication_schedule.medication.add(patient_medication)
        patient_medication_schedule.save()

        return patient_medication_schedule
    
    @classmethod
    def create_by_patient_and_indication(cls, patient, medication_indication):
        from endoreg_db.models import (
            MedicationIndication, PatientMedication,
            Patient
        )

        assert isinstance(medication_indication, MedicationIndication)
        assert isinstance(patient, Patient)
        patient_medication_schedule = cls.objects.create(patient=patient)
        patient_medication_schedule.save()

        patient_medication = PatientMedication.create_by_patient_and_indication(patient, medication_indication)
        patient_medication_schedule.medication.add(patient_medication)
        patient_medication_schedule.save()

        return patient_medication_schedule
    
    def create_patient_medication_from_medication_schedule(
            self,
            medication_schedule,
            medication_indication=None,
            start_date=None,
        ):
        
        from endoreg_db.models import MedicationSchedule, PatientMedication
        from datetime import datetime as dt

        assert isinstance(medication_schedule, MedicationSchedule)

        if not start_date:
            start_date = dt.now()

        drug = medication_schedule.medication
        unit = medication_schedule.unit
        dosage = medication_schedule.dose
        intake_times = medication_schedule.get_intake_times()

        patient_medication = PatientMedication.objects.create(
            patient=self.patient,
            medication=drug,
            medication_indication=medication_indication,
            unit=unit,
            dosage=dosage
        )

        patient_medication.intake_times.set(intake_times)
        patient_medication.save()

        self.medication.add(patient_medication)
        self.save()  

        return patient_medication
    
    
    def get_patient_medication(self):
        return self.medication.all()


from endoreg_db.models import (
    Patient, 
    Center, Disease,
    PatientDisease,
)
    
from datetime import datetime

from django.core.management import call_command
from django.test import TestCase
from io import StringIO
from .conf import (
    TEST_CENTER_NAME,
    TEST_PATIENT_DISEASES_OUTPUT_PATH
)

class TestPatientDiseases(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_disease_data", stdout=out)
        Center.objects.create(name=TEST_CENTER_NAME)

    def test_patient_disease(self):
        patient = Patient.create_generic(center=TEST_CENTER_NAME)

        # get all diseases
        diseases = Disease.objects.all()

        out = ""

        for disease in diseases:
            classifications = disease.get_classifications()
            out += f"{disease.name}:\n"

            for classification in classifications:
                choices = classification.get_choices()
                out += f"\t{classification.name}:\n"

                for choice in choices:

                    _patient_disease = PatientDisease.objects.create(
                        patient=patient,
                        disease=disease,
                        start_date=datetime.now()
                    )

                    out += f"\t\t{choice.name}\n"
                    out += f"\t\t\t{_patient_disease}\n\n"

        with open(TEST_PATIENT_DISEASES_OUTPUT_PATH, "w") as f:
            f.write(out)


        






from django.db import models

class ActiveModelManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)
    
class ActiveModel(models.Model):
    name = models.CharField(max_length=255, unique=True)
    model_meta = models.OneToOneField('ModelMeta', on_delete=models.SET_NULL, blank=True, null=True)



from endoreg_db.models import (
    Patient, 
    Center,
    PatientExamination,
    ExaminationIndication,
    PatientExaminationIndication
)
from .utils import (
    create_patient_with_colonoscopy,
    create_patient_with_endoscopies
)

from django.core.management import call_command
from django.test import TestCase
from io import StringIO
from .conf import (
    TEST_CENTER_NAME,
    TEST_EXAMINATION_NAME_STRINGS,
    TEST_PATIENT_EXAMINATION_OUTPUT_PATH,
    TEST_PATIENT_EXAMINATION_INDICATION_OUTPUT_PATH
)

class TestGeneratePatient(TestCase):
    def setUp(self):
        out = StringIO()
        call_command("load_gender_data", stdout=out)
        call_command("load_unit_data", stdout=out)
        call_command("load_examination_data", stdout=out)
        call_command("load_examination_indication_data", stdout=out)
        Center.objects.create(name=TEST_CENTER_NAME)

        # delete existing test_patient_examination_output_path
        with open(TEST_PATIENT_EXAMINATION_OUTPUT_PATH, "w") as f:
            f.write("")
    
    def test_create_colonoscopy(self):
        patient = Patient.create_generic(center=TEST_CENTER_NAME)

        patient_examination = patient.create_examination()

        assert patient_examination


    def test_create_endoscopies(self):
        # delete all patients and patient examinations
        Patient.objects.all().delete()
        PatientExamination.objects.all().delete()

        patient, patient_examinations = create_patient_with_endoscopies(center_name=TEST_CENTER_NAME)

        examination_name_strings = TEST_EXAMINATION_NAME_STRINGS
        out_str = ""

        for examination_name_str in examination_name_strings:
            patient_examination = PatientExamination.objects.get(
                patient=patient,
                examination__name=examination_name_str
            )

            assert patient_examination

            out_str += f"{patient_examination}\n"
            out_str += f"Age at examination: {patient_examination.get_patient_age_at_examination()}\n\n"

        with open(TEST_PATIENT_EXAMINATION_OUTPUT_PATH, "w") as f:
            f.write(out_str)

    def test_examination_indication(self):
        patient = Patient.create_generic(center=TEST_CENTER_NAME)

        examination_indications = ExaminationIndication.objects.all()

        for examination_indication in examination_indications:
            for examination_indication_choice in examination_indication.get_choices():
                patient_examination, patient_examination_indication = patient.create_examination_by_indication(
                    indication=examination_indication
                )

                assert patient_examination
                assert patient_examination_indication

                patient_examination_indication.indication_choice = examination_indication_choice
                patient_examination_indication.save()

        


        # get all examinations for patient
        patient_examinations = PatientExamination.objects.filter(patient=patient)

        with open(TEST_PATIENT_EXAMINATION_INDICATION_OUTPUT_PATH, "w") as f:
            for patient_examination in patient_examinations:
                f.write(f"{patient_examination}\n")
                
                examination_indications = patient_examination.get_indications()
                assert len(examination_indications) == 1
                examination_indication = examination_indications[0]
                f.write(f"Examination Indication: {examination_indication}\n")

                examination_indication_choices = patient_examination.get_indication_choices()
                assert len(examination_indication_choices) == 1
                examination_indication_choice = examination_indication_choices[0]

                f.write(f"Examination Indication Choice: {examination_indication_choice}\n\n")



    


from .patient import Patient, PatientForm
from .patient_event import PatientEvent
from .patient_disease import PatientDisease
from .patient_lab_sample import PatientLabSample, PatientLabSampleType
from .patient_lab_value import PatientLabValue
from .patient_medication import PatientMedication
from .patient_medication_schedule import PatientMedicationSchedule
from .patient_examination_indication import PatientExaminationIndication
